import time

from pyroute2.netlink.nfnetlink.nftsocket import NFPROTO_IPV4
from pyroute2.nftables.main import NFTables, NFTSetElem


def test_ipv4_addr_set():
    with NFTables(nfgen_family=NFPROTO_IPV4) as nft:
        nft.table("add", name="filter")
        my_set = nft.sets(
            "add",
            table="filter",
            name="test0",
            key_type="ipv4_addr",
            comment="my test set",
            timeout=0,
        )

        # With str
        nft.set_elems(
            "add",
            table="filter",
            set="test0",
            elements={"10.2.3.4", "10.4.3.2"},
        )

        # With NFTSet & NFTSetElem classes
        nft.set_elems(
            "add",
            set=my_set,
            elements={NFTSetElem(value="9.9.9.9", timeout=1000)},
        )

        try:
            assert {
                e.value
                for e in nft.set_elems("get", table="filter", set="test0")
            } == {"10.2.3.4", "10.4.3.2", "9.9.9.9"}
            assert (
                nft.sets("get", table="filter", name="test0").comment
                == b"my test set"
            )

            time.sleep(1.2)
            # timeout for elem 9.9.9.9 (1000ms)
            assert {
                e.value
                for e in nft.set_elems("get", table="filter", set="test0")
            } == {"10.2.3.4", "10.4.3.2"}
        finally:
            nft.sets("del", table="filter", name="test0")
            nft.table("del", name="filter")


def main():
    test_ipv4_addr_set()


if __name__ == "__main__":
    main()



from ctypes import (
    Structure,
    addressof,
    c_int,
    c_ubyte,
    c_ushort,
    c_void_p,
    sizeof,
    string_at,
)
from enum import IntEnum


class sock_filter(Structure):
    _fields_ = [
        ('code', c_ushort),  # u16
        ('jt', c_ubyte),  # u8
        ('jf', c_ubyte),  # u8
        ('k', c_int),  # can be signed or unsigned
    ]


class sock_fprog(Structure):
    _fields_ = [('len', c_ushort), ('filter', c_void_p)]


def compile(code: list[list[int]]):
    ProgramType = sock_filter * len(code)
    program = ProgramType(*[sock_filter(*line) for line in code])
    sfp = sock_fprog(len(code), addressof(program[0]))
    return string_at(addressof(sfp), sizeof(sfp)), program


class BPF(IntEnum):
    '''BPF constants.

    See:
    - https://www.kernel.org/doc/Documentation/networking/filter.txt
    - https://github.com/iovisor/bpf-docs/blob/master/eBPF.md
    '''

    # Operations
    LD = 0x00  # Load
    LDX = 0x01  # Load Index
    ST = 0x02  # Store
    STX = 0x03  # Store Index
    ALU = 0x04  # Arithmetic Logic Unit
    JMP = 0x05  # Jump
    RET = 0x06  # Return
    MISC = 0x07  # Miscellaneous

    # Sizes
    W = 0x00  # Word (4 bytes)
    H = 0x08  # Half-word (2 bytes)
    B = 0x10  # Byte (1 byte)

    # Offsets
    IMM = 0x00  # Immediate
    ABS = 0x20  # Absolute
    IND = 0x40  # Indirect
    MEM = 0x60  # Memory
    LEN = 0x80  # Packet Length
    MSH = 0xA0  # Masked

    # gotos
    JEQ = 0x10  # Jump if Equal
    JGT = 0x20  # Jump if Greater
    JGE = 0x30  # Jump if Greater or Equal
    JSET = 0x40  # Jump if Bit is Set

    # Sources
    K = 0x00  # Constant
    X = 0x08  # Register

    # Operators
    ADD = 0x00
    SUB = 0x10
    MUL = 0x20
    DIV = 0x30
    OR = 0x40
    AND = 0x50
    LSH = 0x60
    RSH = 0x70
    NEG = 0x80
    MOD = 0x90
    XOR = 0xA0
    MOV = 0xB0
    ARSH = 0xC0
    END = 0xD0



class cmsg(dict):
    def __init__(self, target, payload=None):
        self['header'] = {'target': target}
        self.payload = payload


class cmsg_event(cmsg):
    pass


class cmsg_failed(cmsg):
    pass


class cmsg_sstart(cmsg):
    pass



import errno
import json
import shutil
import socket
import ssl
import subprocess

from pyroute2.netlink import nlmsg
from pyroute2.netlink.exceptions import NetlinkError


class probe_msg(nlmsg):
    '''
    Fake message type to represent network probe info.

    This is a prototype, the NLA layout is subject to change without
    notification.
    '''

    __slots__ = ()
    prefix = 'PROBE_'

    fields = (('family', 'B'), ('proto', 'B'), ('port', 'H'), ('dst_len', 'I'))

    nla_map = (
        ('PROBE_UNSPEC', 'none'),
        ('PROBE_KIND', 'asciiz'),
        ('PROBE_STDOUT', 'asciiz'),
        ('PROBE_STDERR', 'asciiz'),
        ('PROBE_SRC', 'asciiz'),
        ('PROBE_DST', 'asciiz'),
        ('PROBE_NUM', 'uint8'),
        ('PROBE_TIMEOUT', 'uint8'),
        ('PROBE_HOSTNAME', 'asciiz'),
        ('PROBE_SSL_VERIFY', 'uint8'),
        ('PROBE_SSL_VERSION', 'asciiz'),
        ('PROBE_SSL_CERT_JSON', 'asciiz'),
        ('PROBE_SSL_CERT_DER', 'cdata'),
    )


def probe_ping(msg):
    num = msg.get('num')
    timeout = msg.get('timeout')
    dst = msg.get('dst')
    kind = msg.get('kind')
    args = [shutil.which(kind), '-c', f'{num}', '-W', f'{timeout}', f'{dst}']
    if args[0] is None:
        raise NetlinkError(errno.ENOENT, 'probe not found')

    process = subprocess.Popen(
        args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    try:
        out, err = process.communicate(timeout=timeout)
        if out:
            msg['attrs'].append(['PROBE_STDOUT', out])
        if err:
            msg['attrs'].append(['PROBE_STDERR', err])
    except subprocess.TimeoutExpired:
        process.terminate()
        raise NetlinkError(errno.ETIMEDOUT, 'timeout expired')
    finally:
        process.stdout.close()
        process.stderr.close()
        return_code = process.wait()
    if return_code != 0:
        raise NetlinkError(errno.EHOSTUNREACH, 'probe failed')


def probe_tcp(msg, close=True):
    timeout = msg.get('timeout')
    dst = msg.get('dst')
    port = msg.get('port')
    connection = None
    try:
        connection = socket.create_connection((dst, port), timeout=timeout)
    except ConnectionRefusedError:
        raise NetlinkError(errno.ECONNREFUSED, 'connection refused')
    except TimeoutError:
        raise NetlinkError(errno.ETIMEDOUT, 'timeout expired')
    except Exception:
        raise NetlinkError(errno.ECOMM, 'probe failed')
    finally:
        if close and connection is not None:
            connection.close()
    return connection


def probe_ssl(msg):
    hostname = msg.get('hostname') or msg.get('dst')
    context = ssl.create_default_context()
    context.verify_mode = msg.get('ssl_verify', ssl.CERT_REQUIRED)
    with probe_tcp(msg, close=False) as connection:
        try:
            with context.wrap_socket(
                connection, server_hostname=hostname
            ) as ssl_wrap:
                version = ssl_wrap.version()
                peer_cert_json = ssl_wrap.getpeercert(binary_form=False)
                peer_cert_der = ssl_wrap.getpeercert(binary_form=True)
                if peer_cert_json is not None:
                    msg['attrs'].append(
                        ['PROBE_SSL_CERT_JSON', json.dumps(peer_cert_json)]
                    )
                if peer_cert_der is not None:
                    msg['attrs'].append(['PROBE_SSL_CERT_DER', peer_cert_der])
                if version is not None:
                    msg['attrs'].append(['PROBE_SSL_VERSION', version])
        except ssl.SSLError as e:
            code = errno.EPROTO
            if e.reason == 'UNSUPPORTED_PROTOCOL':
                code = errno.EPROTONOSUPPORT
            elif e.reason == 'CERTIFICATE_VERIFY_FAILED':
                code = errno.EACCES
            raise NetlinkError(code, e.strerror)
        except Exception:
            raise NetlinkError(errno.ECOMM, 'probe failed')


def proxy_newprobe(msg):
    kind = msg.get('kind')
    if kind.startswith('ping'):
        probe_ping(msg)
    elif kind == 'tcp':
        probe_tcp(msg)
    elif kind == 'ssl':
        probe_ssl(msg)
    else:
        raise NetlinkError(errno.ENOTSUP, 'probe type not supported')
    msg.reset()
    msg.encode()
    return msg.data



from socket import AF_INET, inet_ntop, inet_pton

from pyroute2.dhcp import Policy, dhcpmsg, option
from pyroute2.dhcp.enums.dhcp import Option


class dhcp4msg(dhcpmsg):
    #
    # https://www.ietf.org/rfc/rfc2131.txt
    #
    fields = (
        ('op', 'uint8', 1),  # request
        ('htype', 'uint8', 1),  # ethernet
        ('hlen', 'uint8', 6),  # ethernet addr len
        ('hops', 'uint8'),
        ('xid', 'be32'),
        ('secs', 'be16'),
        ('flags', 'be16'),
        ('ciaddr', 'ip4addr'),
        ('yiaddr', 'ip4addr'),
        ('siaddr', 'ip4addr'),
        ('giaddr', 'ip4addr'),
        ('chaddr', 'l2paddr'),
        ('sname', '64s'),
        ('file', '128s'),
        ('cookie', '4s', b'c\x82Sc'),
    )
    #
    # https://www.ietf.org/rfc/rfc2132.txt
    #
    options = (
        # TODO: add & test more options
        (Option.PAD, 'none'),
        (Option.SUBNET_MASK, 'ip4addr'),
        (Option.TIME_OFFSET, 'be32'),
        (Option.ROUTER, 'ip4list'),
        (Option.TIME_SERVER, 'ip4list'),
        (Option.IEN_NAME_SERVER, 'ip4list'),
        (Option.NAME_SERVER, 'ip4list'),
        (Option.LOG_SERVER, 'ip4list'),
        (Option.COOKIE_SERVER, 'ip4list'),
        (Option.LPR_SERVER, 'ip4list'),
        (Option.IMPRESS_SERVER, 'ip4list'),
        (Option.RESOURCE_LOCATION_SERVER, 'ip4list'),
        (Option.HOST_NAME, 'string'),
        # in multiples of 512 bytes
        (Option.BOOT_FILE_SIZE, 'be16'),
        (Option.MERIT_DUMP_FILE, 'string'),
        (Option.DOMAIN_NAME, 'string'),
        (Option.SWAP_SERVER, 'ip4addr'),
        (Option.ROOT_PATH, 'string'),
        (Option.EXTENSIONS_PATH, 'string'),
        (Option.IP_FORWARDING, 'bool'),
        (Option.NON_LOCAL_SOURCE_ROUTING, 'bool'),
        # TODO: dict of ip/mask pairs
        # (Option.POLICY_FILTER, '?'),
        # minimum legal value is 567 according to rfc
        (Option.MAX_DATAGRAM_REASSEMBLY, 'be16'),
        (Option.DEFAULT_TTL, 'uint8'),
        (Option.PATH_MTU_AGING_TIMEOUT, 'be32'),
        # TODO: list of be16
        # (Option.PATH_MTU_PLATEAU_TABLE, '?'),
        (Option.INTERFACE_MTU, 'be16'),
        (Option.ALL_SUBNETS_LOCAL, 'bool'),
        (Option.BROADCAST_ADDRESS, 'ip4addr'),
        (Option.PERFORM_MASK_DISCOVERY, 'bool'),
        (Option.MASK_SUPPLIER, 'bool'),
        (Option.PERFORM_ROUTER_DISCOVERY, 'bool'),
        (Option.ROUTER_SOLICITATION_ADDRESS, 'ip4addr'),
        # TODO: dict of ip addrs
        # (Option.STATIC_ROUTE, '?')
        (Option.TRAILER_ENCAPSULATION, 'bool'),
        (Option.ARP_CACHE_TIMEOUT, 'be32'),
        (Option.ETHERNET_ENCAPSULATION, 'bool'),
        (Option.TCP_DEFAULT_TTL, 'uint8'),
        (Option.TCP_KEEPALIVE_INTERVAL, 'be32'),
        (Option.TCP_KEEPALIVE_GARBAGE, 'bool'),
        (Option.NIS_DOMAIN, 'string'),
        (Option.NDS_SERVERS, 'ip4list'),
        (Option.NTP_SERVERS, 'ip4list'),
        (Option.VENDOR_SPECIFIC_INFORMATION, 'string'),
        (Option.NETBIOS_NAME_SERVER, 'ip4list'),
        (Option.NETBIOS_DDG_SERVER, 'ip4list'),
        # 1=B, 2=P, 4=M, 8=H, probably obsolete & unused
        (Option.NETBIOS_NODE_TYPE, 'uint8'),
        (Option.NETBIOS_SCOPE, 'string'),
        (Option.X_WINDOW_FONT_SERVER, 'ip4list'),
        (Option.X_WINDOW_DISPLAY_MANAGER, 'ip4list'),
        (Option.REQUESTED_IP, 'ip4addr'),
        (Option.LEASE_TIME, 'sbe32'),
        # 1: options in file, 2: options in sname, 4, both
        (Option.OPTION_OVERLOAD, 'uint8'),
        (Option.TFTP_SERVER_NAME, 'string'),
        (Option.BOOTFILE_NAME, 'string'),
        (Option.MESSAGE_TYPE, 'message_type'),
        (Option.SERVER_ID, 'ip4addr'),
        (Option.PARAMETER_LIST, 'array8'),
        (Option.MESSAGE, 'string'),
        # minimum value: 576 bytes
        (Option.MAX_MSG_SIZE, 'be16'),
        (Option.RENEWAL_TIME, 'sbe32'),
        (Option.REBINDING_TIME, 'sbe32'),
        (Option.VENDOR_ID, 'string'),
        (Option.CLIENT_ID, 'client_id'),
        (Option.END, 'none'),
    )

    class ip4addr(option):
        policy = Policy(
            format='4s',
            encode=lambda x: inet_pton(AF_INET, x),
            decode=lambda x: inet_ntop(AF_INET, x),
        )

    class ip4list(option):
        policy = Policy(
            format='string',
            encode=lambda x: b''.join([inet_pton(AF_INET, i) for i in x]),
            decode=lambda x: [
                inet_ntop(AF_INET, x[i * 4 : i * 4 + 4])
                for i in range(len(x) // 4)
            ],
        )



from pyroute2.netlink.rt_files import EmatchMapFile
from pyroute2.netlink.rtnl.tcmsg import em_cmp, em_ipset, em_meta

plugins = {
    # 0: em_container,
    1: em_cmp,
    # 2: em_nbyte,
    # 3: em_u32,
    4: em_meta,
    # 5: em_text,
    # 6: em_vlan,
    # 7: em_canid,
    8: em_ipset,
    # 9: em_ipt,
}

plugins_translate = {
    'container': 0,
    'cmp': 1,
    'nbyte': 2,
    'u32': 3,
    'meta': 4,
    'text': 5,
    'vlan': 6,
    'canid': 7,
    'ipset': 8,
    'ipt': 9,
}

TCF_EM_REL_END = 0
TCF_EM_REL_AND = 1
TCF_EM_REL_OR = 2
TCF_EM_INVERSE_MASK = 4

RELATIONS_DICT = {
    'and': TCF_EM_REL_AND,
    'AND': TCF_EM_REL_AND,
    '&&': TCF_EM_REL_AND,
    'or': TCF_EM_REL_OR,
    'OR': TCF_EM_REL_OR,
    '||': TCF_EM_REL_OR,
}


class nla_plus_tcf_ematch_opt(object):
    @staticmethod
    def parse_ematch_options(self, *argv, **kwarg):
        if 'kind' not in self:
            raise ValueError('ematch requires "kind" parameter')

        kind = self['kind']
        if kind in plugins:
            ret = plugins[kind].data(data=argv[0])
            ret.decode()
            return ret
        return self.hex


def get_ematch_parms(kwarg):
    if 'kind' not in kwarg:
        raise ValueError('ematch requires "kind" parameter')

    if kwarg['kind'] in plugins:
        return plugins[kwarg['kind']].get_parameters(kwarg)
    else:
        return []


def get_tcf_ematches(kwarg):
    ret = {'attrs': []}
    matches = []
    header = {'nmatches': 0, 'progid': 0}

    # Get the number of expressions
    expr_count = len(kwarg['match'])
    header['nmatches'] = expr_count

    # Load plugin and transfer data
    for i in range(0, expr_count):
        match = {'matchid': 0, 'kind': None, 'flags': 0, 'pad': 0, 'opt': None}

        cur_match = kwarg['match'][i]

        # Translate string kind into numeric kind
        try:
            kind = plugins_translate[cur_match['kind']]
        except KeyError:
            kind = EmatchMapFile().get_rt_id(cur_match['kind'])
        match['kind'] = kind
        data = plugins[kind].data()
        data.setvalue(cur_match)
        data.encode()

        # Add ematch encoded data
        match['opt'] = data.data

        # Safety check
        if i == expr_count - 1 and 'relation' in cur_match:
            raise ValueError('Could not set a relation to the last expression')

        if i < expr_count - 1 and 'relation' not in cur_match:
            raise ValueError(
                'You must specify a relation for every expression'
                ' except the last one'
            )

        # Set relation to flags
        if 'relation' in cur_match:
            relation = cur_match['relation']
            if relation in RELATIONS_DICT:
                match['flags'] |= RELATIONS_DICT.get(relation)
            else:
                raise ValueError('Unknown relation {0}'.format(relation))
        else:
            match['flags'] = TCF_EM_REL_END

        # Handle inverse flag
        if 'inverse' in cur_match:
            if cur_match['inverse']:
                match['flags'] |= TCF_EM_INVERSE_MASK

        # Append new match to list of matches
        matches.append(match)

    ret['attrs'].append(['TCA_EMATCH_TREE_HDR', header])
    ret['attrs'].append(['TCA_EMATCH_TREE_LIST', matches])

    return ret



import json

import pytest

from pyroute2 import IPRoute
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
from pyroute2.requests.link import LinkFieldFilter, LinkIPRouteFilter
from pyroute2.requests.main import RequestProcessor

with open('test_unit/test_iproute_match/links.dump', 'r') as f:
    ifinfmsg_sample = [ifinfmsg().load(x) for x in json.load(f)]
for msg in ifinfmsg_sample:
    msg.reset()
    msg.encode()
    msg.decode()


@pytest.fixture
def ipr():
    with IPRoute() as iproute:
        yield iproute


@pytest.mark.parametrize(
    'spec,query,result',
    (
        ({'ifname': 'lo'}, ('stats64', 'rx_packets'), 120),
        ({'ifname': 'lo'}, ('af_spec', 'af_inet', 'forwarding'), 1),
        ({'ifname': 'wl0'}, ('num_rx_queues',), 1),
        ({'ifname': 'wl0'}, ('qdisc',), 'noqueue'),
        ({'ifname': 'wl0'}, ('stats64', 'rx_packets'), 835511),
        (
            {'ifname': 'wl0'},
            ('af_spec', 'af_inet6', 'inet6_flags'),
            2147483648,
        ),
        (
            {'ifname': 'wl0'},
            ('af_spec', 'af_inet6', 'inet6_conf', 'temp_preferred_lft'),
            86400,
        ),
        ({'parent_dev_name': '0000:03:00.0'}, ('ifname',), 'wl0'),
        ({'kind': 'bridge', 'br_forward_delay': 1500}, ('ifname',), 'br0'),
        (
            {'ifname': 'br0'},
            ('linkinfo', 'data', 'br_group_addr'),
            '01:80:c2:00:00:00',
        ),
    ),
    ids=[
        'lo:stats64/rx_packets',
        'lo:af_spec/af_inet/forwarding',
        'wl0:num_rx_queues',
        'wl0:qdisc',
        'wl0:stats64/rx_packets',
        'wl0:af_spec/af_inet6/inet6_flags',
        'wl0:af_spec/af_inet6/inet6_conf/temp_preferred_lft',
        'parent_dev_name(...) => wl0',
        'br_forward_delay(...) => br0',
        'br0:linkinfo/data/br_group_addr',
    ],
)
def test_get_leaf(ipr, spec, query, result):
    spec = RequestProcessor(context=spec, prime=spec)
    spec.add_filter(LinkFieldFilter())
    spec.add_filter(LinkIPRouteFilter('dump'))
    spec.finalize()

    msg = tuple(ipr.filter_messages(spec, ifinfmsg_sample))
    assert len(msg) == 1
    assert msg[0].get_nested(*query) == result



'''
DHCP protocol
=============

The DHCP implementation here is far from complete, but
already provides some basic functionality. Later it will
be extended with IPv6 support and more DHCP options
will be added.

Right now it can be interesting mostly to developers,
but not users and/or system administrators. So, the
development hints first.

The packet structure description is intentionally
implemented as for netlink packets. Later these two
parsers, netlink and generic, can be merged, so the
syntax is more or less compatible.

Packet fields
-------------

There are two big groups of items within any DHCP packet.
First, there are BOOTP/DHCP packet fields, they're defined
with the `fields` attribute::

    class dhcp4msg(msg):
        fields = ((name, format, policy),
                  (name, format, policy),
                  ...
                  (name, format, policy))

The `name` can be any literal. Format should be specified
as for the struct module, like `B` for `uint8`, or `i` for
`int32`, or `>Q` for big-endian uint64. There are also
aliases defined, so one can write `uint8` or `be16`, or
like that. Possible aliases can be seen in the
`pyroute2.protocols` module.

The `policy` is a bit complicated. It can be a number or
literal, and it will mean that it is a default value, that
should be encoded if no other value is given.

But when the `policy` is a dictionary, it can contain keys
as follows::

    'l2addr': {'format': '6B',
               'decode': ...,
               'encode': ...}

Keys `encode` and `decode` should contain filters to be used
in decoding and encoding procedures. The encoding filter
should accept the value from user's definition and should
return a value that can be packed using `format`. The decoding
filter should accept a value, decoded according to `format`,
and should return value that can be used by a user.

The `struct` module can not decode IP addresses etc, so they
should be decoded as `4s`, e.g. Further transformation from
4 bytes string to a string like '10.0.0.1' performs the filter.

DHCP options
------------

DHCP options are described in a similar way::

    options = ((code, name, format),
               (code, name, format),
               ...
               (code, name, format))

Code is a `uint8` value, name can be any string literal. Format
is a string, that must have a corresponding class, inherited from
`pyroute2.dhcp.option`. One can find these classes in
`pyroute2.dhcp` (more generic) or in `pyroute2.dhcp.dhcp4msg`
(IPv4-specific). The option class must reside within dhcp message
class.

Every option class can be decoded in two ways. If it has fixed
width fields, it can be decoded with ordinary `msg` routines, and
in this case it can look like that::

    class client_id(option):
        fields = (('type', 'uint8'),
                  ('key', 'l2addr'))

If it must be decoded by some custom rules, one can define the
policy just like for the fields above::

    class array8(option):
        policy = {'format': 'string',
                  'encode': lambda x: array('B', x).tobytes(),
                  'decode': lambda x: array('B', x).tolist()}

In the corresponding modules, like in `pyroute2.dhcp.dhcp4msg`,
one can define as many custom DHCP options, as one need. Just
be sure, that they are compatible with the DHCP server and all
fit into 1..254 (`uint8`) -- the 0 code is used for padding and
the code 255 is the end of options code.
'''

import logging
import struct
from array import array
from typing import ClassVar, NamedTuple, Optional, TypedDict, TypeVar, Union

from pyroute2.protocols import Policy, _decode_mac, _encode_mac, msg

from .enums.dhcp import MessageType, Option

LOG = logging.getLogger(__name__)


# hack because Self is not supported in py39
_dhcpmsgSelf = TypeVar('_dhcpmsgSelf', bound='dhcpmsg')
_optionSelf = TypeVar('_optionSelf', bound='option')


class ClientId(TypedDict):
    '''A dict with 'type' and 'key' keys.

    The types stores the client id type, and the key stores the value.
    See RFC for their meaning.
    '''

    type: int
    key: Union[bytes, str]


def _decode_client_id(value: bytes) -> ClientId:
    '''Decode a raw client id option into a dict with type and key.

    If the type is 1, the key is decoded as a mac address,
    otherwise it's just the raw bytes.
    '''
    type_ = value[0]
    key: Union[bytes, str] = value[1:]

    if type_ == 1:
        # ethernet
        assert isinstance(key, bytes)
        key = _decode_mac(value=key)
    return ClientId(type=type_, key=key)


def _encode_client_id(value: ClientId) -> bytes:
    '''Encode a client_id dict into bytes.'''
    type_ = value['type']
    key = value['key']
    if type_ == 1:
        assert isinstance(key, str), 'client_id must be a mac str when type=1'
        key = bytes(_encode_mac(key))
    assert isinstance(key, bytes), 'client_id must be bytes'
    return struct.pack('B', type_) + key


class option(msg):
    policy: ClassVar[Optional[Policy]] = None

    def __init__(self, content=None, buf=b'', offset=0, value=None, code=0):
        msg.__init__(
            self, content=content, buf=buf, offset=offset, value=value
        )
        self.code = code

    @property
    def length(self) -> Optional[int]:
        if self.data_length is None:
            return None
        if self.data_length == 0:
            return 1
        else:
            return self.data_length + 2

    def encode(self: _optionSelf) -> _optionSelf:
        # pack code
        self.buf += struct.pack('B', self.code)
        if self.code in (0, 255):
            return self
        # save buf
        save = self.buf
        self.buf = b''
        # pack data into the new buf
        if self.policy is not None:
            value = self.policy.encode(self.value)
            if self.policy.format == 'string':
                fmt = '%is' % len(value)
                if isinstance(value, list):
                    # Byte strings can be provided as a list of bytes
                    value = bytes(value)
            else:
                fmt = self.policy.format
            if isinstance(value, str):
                value = value.encode('utf-8')
            self.buf = struct.pack(fmt, value)
        else:
            msg.encode(self)
        # get the length
        data = self.buf
        self.buf = save
        self.buf += struct.pack('B', len(data))
        # attach the packed data
        self.buf += data
        return self

    def decode(self: _optionSelf) -> _optionSelf:
        try:
            self.data_length = struct.unpack(
                'B', self.buf[self.offset + 1 : self.offset + 2]
            )[0]
        except struct.error as err:
            raise ValueError(
                f'Cannot decode length for DHCP option {self.code}: {err}'
            )
        if self.policy is not None:
            if self.policy.format == 'string':
                fmt = '%is' % self.data_length
            else:
                fmt = self.policy.format
            try:
                value = struct.unpack(
                    fmt,
                    self.buf[
                        self.offset + 2 : self.offset + 2 + self.data_length
                    ],
                )
            except struct.error as err:
                raise ValueError(
                    f'Cannot decode option {self.code} '
                    f'as {self.policy.format}: {err}'
                )
            if len(value) == 1:
                value = value[0]
            value = self.policy.decode(value)
            if self.policy.format == 'string':
                if isinstance(value, bytes):
                    try:
                        # Try to decode as a string
                        value = value.decode()
                    except ValueError:
                        pass
                if isinstance(value, str):
                    # Strip trailing zeroes for strings
                    value = value.rstrip("\x00")
            if isinstance(value, bytes):
                # Turn bytes to lists of bytes so they're JSON-encodable
                value = list(value)
            self.value = value
        else:
            # remember current offset as msg.decode() will advance it
            offset = self.offset
            # move past the code and option length bytes so that msg.decode()
            # starts parsing at the right location
            self.offset += 2
            msg.decode(self)
            # restore offset so that dhcpmsg.decode() advances it correctly
            self.offset = offset

        return self


class CodeMapping(NamedTuple):
    name: str
    code: int
    format: str


class dhcpmsg(msg):
    options: ClassVar[tuple[tuple[Option, str], ...]] = ()

    def __init__(self, content=None, buf=b'', offset=0, value=None) -> None:
        super().__init__(content, buf, offset, value)
        self._encode_map: dict[str, CodeMapping] = {}
        self._decode_map: dict[int, CodeMapping] = {}
        self._register_options()

    def _register_options(self) -> None:
        for code, fmt in self.options:
            name = code.name.lower()
            self._decode_map[code] = self._encode_map[name] = CodeMapping(
                name=name, code=code, format=fmt
            )

    def decode(self: _dhcpmsgSelf) -> _dhcpmsgSelf:
        msg.decode(self)
        self['options'] = {}
        while self.offset < len(self.buf):
            code: int = struct.unpack(
                'B', self.buf[self.offset : self.offset + 1]
            )[0]
            if code == Option.PAD:
                self.offset += 1
                continue
            if code == Option.END:
                return self
            if code in self._decode_map:
                # use the known decoded & name
                option_class = getattr(self, self._decode_map[code].format)
                optname = self._decode_map[code].name
            else:
                # code is unknown
                # if we know this option number, get the value as a list
                # even if we don't know how to parse it
                option_class = self.array8
                try:
                    # if we know this option name, use it
                    code = Option(code)
                    optname = code.name.lower()
                except ValueError:
                    optname = f"option{code}"

            option = option_class(buf=self.buf, offset=self.offset, code=code)
            try:
                option.decode()
            except ValueError as err:
                # FIXME: maybe we would like the raw option data
                # if we can't decode it ? but that would complicate typing
                LOG.error("%s", err)
                break
            self.offset += option.length
            if option.value is not None:
                value = option.value
            else:
                value = option
            self['options'][optname] = value
        return self

    def encode(self: _dhcpmsgSelf) -> _dhcpmsgSelf:
        msg.encode(self)
        # put message type
        options = self.get('options') or {'message_type': MessageType.DISCOVER}

        self.buf += (
            self.uint8(code=Option.MESSAGE_TYPE, value=options['message_type'])
            .encode()
            .buf
        )
        for name, value in options.items():
            if name == 'message_type':
                continue
            if (code_mapping := self._encode_map.get(name)) is None:
                continue

            # name is known, ok
            option_class = getattr(self, code_mapping.format)
            if isinstance(value, dict):
                option = option_class(value, code=code_mapping.code)
            else:
                option = option_class(code=code_mapping.code, value=value)
            self.buf += option.encode().buf

        self.buf += self.none(code=Option.END).encode().buf
        return self

    class none(option):
        pass

    class be16(option):
        policy = Policy(format='>H')

    class be32(option):
        policy = Policy(format='>I')

    class sbe32(option):
        policy = Policy(format='>i')

    class uint8(option):
        policy = Policy(format='B')

    class string(option):
        policy = Policy(format='string')

    class array8(option):
        policy = Policy(
            format='string',
            encode=lambda x: array('B', x).tobytes(),
            decode=lambda x: array('B', x).tolist(),
        )

    class client_id(option):
        policy = Policy(
            format='string', encode=_encode_client_id, decode=_decode_client_id
        )

        def __init__(
            self, content=None, buf=b'', offset=0, value=None, code=0
        ):
            # FIXME: we have to override value w/ content here, because
            # when trying to encode an option, if it's a dict, its value is
            # set to None.
            super().__init__(content, buf, offset, value=content, code=code)

    class message_type(option):
        policy = Policy(format='B', decode=MessageType)

    class bool(option):
        policy = Policy(format='B', encode=int, decode=bool)



from enum import IntEnum


class MessageType(IntEnum):
    '''DHCP message types, see RFC 2131 table 2.'''

    DISCOVER = 1
    OFFER = 2
    REQUEST = 3
    DECLINE = 4
    ACK = 5
    NAK = 6
    RELEASE = 7
    INFORM = 8


class Option(IntEnum):
    '''DHCP options and parameters.

    These constants are used for two purposes:

    - Requesting configuration values from the DHCP server.
      In this case, they're passed in the `PARAMETER_LIST` request option.
      Not all values are valid for parameter requests:
        - `PAD` is only used as padding when decoding server-sent options,
        - `PARAMETER_LIST` is only sent from client and cannot be requested,
        - `END` is used as a marker when decoding server-sent options.

    - Reading responses.
      `Option` numbers are used to parse DHCP options sent by the server.
      They are translated to their name by the client and stored in leases.
    '''

    PAD = 0
    SUBNET_MASK = 1
    TIME_OFFSET = 2
    ROUTER = 3
    TIME_SERVER = 4
    IEN_NAME_SERVER = 5  # prehistoric DNS
    # this should be DOMAIN_NAME_SERVER but it's often used & shorted this way
    NAME_SERVER = 6  # plain old DNS
    LOG_SERVER = 7
    COOKIE_SERVER = 8
    LPR_SERVER = 9
    IMPRESS_SERVER = 10
    RESOURCE_LOCATION_SERVER = 11
    HOST_NAME = 12
    BOOT_FILE_SIZE = 13
    MERIT_DUMP_FILE = 14
    DOMAIN_NAME = 15
    SWAP_SERVER = 16
    ROOT_PATH = 17
    EXTENSIONS_PATH = 18
    IP_FORWARDING = 19
    NON_LOCAL_SOURCE_ROUTING = 20
    POLICY_FILTER = 21
    MAX_DATAGRAM_REASSEMBLY = 22
    DEFAULT_TTL = 23
    PATH_MTU_AGING_TIMEOUT = 24
    PATH_MTU_PLATEAU_TABLE = 25
    INTERFACE_MTU = 26
    ALL_SUBNETS_LOCAL = 27
    BROADCAST_ADDRESS = 28
    PERFORM_MASK_DISCOVERY = 29
    MASK_SUPPLIER = 30
    PERFORM_ROUTER_DISCOVERY = 31
    ROUTER_SOLICITATION_ADDRESS = 32
    STATIC_ROUTE = 33
    TRAILER_ENCAPSULATION = 34
    ARP_CACHE_TIMEOUT = 35
    ETHERNET_ENCAPSULATION = 36
    TCP_DEFAULT_TTL = 37
    TCP_KEEPALIVE_INTERVAL = 38
    TCP_KEEPALIVE_GARBAGE = 39
    NIS_DOMAIN = 40
    NIS_SERVERS = 41
    NTP_SERVERS = 42
    VENDOR_SPECIFIC_INFORMATION = 43
    NETBIOS_NAME_SERVER = 44
    NETBIOS_DDG_SERVER = 45
    NETBIOS_NODE_TYPE = 46
    NETBIOS_SCOPE = 47
    X_WINDOW_FONT_SERVER = 48
    X_WINDOW_DISPLAY_MANAGER = 49
    REQUESTED_IP = 50
    LEASE_TIME = 51
    OPTION_OVERLOAD = 52
    MESSAGE_TYPE = 53
    SERVER_ID = 54
    PARAMETER_LIST = 55
    MESSAGE = 56
    MAX_MSG_SIZE = 57
    RENEWAL_TIME = 58
    REBINDING_TIME = 59
    VENDOR_ID = 60
    CLIENT_ID = 61
    NETWARE_IP_DOMAIN = 62
    NETWARE_IP_OPTION = 63
    NIS_PLUS_DOMAIN = 64
    NIS_PLUS_SERVERS = 65
    TFTP_SERVER_NAME = 66
    BOOTFILE_NAME = 67
    MOBILE_IP_HOME_AGENT = 68
    SMTP_SERVER = 69
    POP3_SERVER = 70
    NNTP_SERVER = 71
    DEFAULT_WWW_SERVER = 72
    DEFAULT_FINGER_SERVER = 73
    DEFAULT_IRC_SERVER = 74
    STREETTALK_SERVER = 75
    STDA_SERVER = 76
    USER_CLASS_INFORMATION = 77
    SLP_DIRECTORY_AGENT = 78
    SLP_SERVICE_SCOPE = 79
    RAPID_COMMIT = 80
    CLIENT_FQDN = 81
    RELAY_AGENT_INFORMATION = 82
    INTERNET_STORAGE_NAME_SERVICE = 83
    # 84 is unused
    NDS_SERVERS = 85
    NDS_TREE_NAME = 86
    NDS_CONTEXT = 87
    BCMCS_CONTROLLER_DOMAIN = 88
    BCMS_CONTROLLER_IP = 89
    AUTHENTICATION = 90
    LQ_CLIENT_LAST_TRANSACTION = 91
    LQ_ASSOCIATED_IP = 92
    CLIENT_SYSTEM_ARCHITECTURE_TYPE = 93
    CLIENT_NETWORK_INTERFACE_IDENTIFIER = 94
    LDAP_SERVERS = 95
    # 96 is unused
    PXE_CLIENT_ID = 97
    OPEN_GROUP_USER_AUTH = 98
    GEOCONF_CIVIC = 99
    POSIX_TIMEZONE = 100
    TZDB_TIMEZONE = 101
    # 102 to 107 are unused
    IPV6_ONLY_PREFERRED = 108
    DHCP4TO6_S46_SADDR = 109
    # 110 & 111 are unused
    NETINFO_PARENT_SERVER_ADDR = 112
    NETINFO_PARENT_SERVER_TAG = 113
    DHCP_CAPTIVE_PORTAL = 114
    # 115 is unused
    AUTO_CONFIGURE = 116
    NAME_SERVICE_SEARCH = 117
    SUBNET_SELECTION = 118
    DOMAIN_SEARCH = 119
    SIP_SERVERS = 120
    CLASSLESS_STATIC_ROUTE = 121
    CABLELABS_CLIENT_CONFIG = 122
    GEOCONF = 123
    VENDOR_IDENTIFYING_CLASS = 124
    VENDOR_IDENTIFYING_OPTS = 125
    # 126 & 127 are unused
    MCNS_SECURITY_SERVER = 128
    # 129 to 135 are unused
    PANA_AGENT = 136
    LOST_SERVER = 137
    CAPWAP_AC_V4 = 138
    MOS_ADDR = 139
    MOS_FQDN = 140
    SIP_UA_CS_DOMAINS = 141
    ANDSF_V4 = 142
    SZTP_REDIRECT = 143
    GEOLOC = 144
    FORCERENEW_NONCE_CAPABLE = 145
    RDNSS_SELECTION = 146
    DOTS_REFERENCE_IDENTIFIER = 147
    DOTS_ADDR = 148
    # 149 is unused
    TFTP_SERVER_ADDR = 150
    STATUS_CODE = 151
    BASE_TIME = 152
    STATE_START_TIME = 153
    QUERY_START_TIME = 154
    QUERY_END_TIME = 155
    DHCP_STATE = 156
    DATA_SOURCE = 157
    PCP_SERVER = 158
    PORT_PARAMS = 159
    OLD_CAPTIVE_PORTAL = 160
    MUD_URL = 161
    CISCO_REQUESTED_HOST_NAME = 162
    CISCO_LAST_TRANSACTION_TIME = 163
    # 164 to 184 are unused
    VPN_ID = 185
    # 186 to 207 are unused
    PXELINUX_MAGIC = 208
    PXELINUX_CONFIG_FILE = 209
    PXELINUX_PATH_PREFIX = 210
    PXELINUX_REBOOT_TIME = 211
    DHCP_6RD = 212
    ACCESS_DOMAIN_NAME = 213
    # 214 to 219 seem unused
    SUBNET_ALLOC = 220
    VIRTUAL_SUBNET_SELECTION = 221
    PRIVATE_CLASSIC_ROUTE_MS = 249
    CISCO_AUTO_CONFIGURE = 251
    PRIVATE_PROXY_AUTODISCOVERY = 252
    END = 255



# -*- coding: utf-8 -*-
'''
Common utilities
'''
import errno
import io
import os
import socket
import struct
import threading
import time
import types
from functools import partial
from typing import Any, Callable, Literal, Optional, TypeVar, Union

basestring = (str, bytes)
file = io.BytesIO

AF_MPLS = 28
_uuid32 = 0  # (singleton) the last uuid32 value saved to avoid collisions
_uuid32_lock = threading.Lock()
USE_DEFAULT_TIMEOUT = -1

size_suffixes = {
    'b': 1,
    'k': 1024,
    'kb': 1024,
    'm': 1024 * 1024,
    'mb': 1024 * 1024,
    'g': 1024 * 1024 * 1024,
    'gb': 1024 * 1024 * 1024,
    'kbit': 1024 / 8,
    'mbit': 1024 * 1024 / 8,
    'gbit': 1024 * 1024 * 1024 / 8,
}


time_suffixes = {
    's': 1,
    'sec': 1,
    'secs': 1,
    'ms': 1000,
    'msec': 1000,
    'msecs': 1000,
    'us': 1000000,
    'usec': 1000000,
    'usecs': 1000000,
}

rate_suffixes = {
    'bit': 1,
    'Kibit': 1024,
    'kbit': 1000,
    'mibit': 1024 * 1024,
    'mbit': 1000000,
    'gibit': 1024 * 1024 * 1024,
    'gbit': 1000000000,
    'tibit': 1024 * 1024 * 1024 * 1024,
    'tbit': 1000000000000,
    'Bps': 8,
    'KiBps': 8 * 1024,
    'KBps': 8000,
    'MiBps': 8 * 1024 * 1024,
    'MBps': 8000000,
    'GiBps': 8 * 1024 * 1024 * 1024,
    'GBps': 8000000000,
    'TiBps': 8 * 1024 * 1024 * 1024 * 1024,
    'TBps': 8000000000000,
}


##
# General purpose
#
class Namespace:
    def __init__(self, parent, override=None):
        raise NotImplementedError()


def _no_change(s: str) -> str:
    return s


def _default_normalize(s: str, prefix: str) -> str:
    return s[len(prefix) :].lower()


def map_namespace(
    prefix: str,
    ns: dict[str, Any],
    normalize: Union[None, Literal[True], Callable[[str], str]] = None,
) -> tuple[dict[str, int], dict[int, str]]:
    '''
    Take the namespace prefix, list all constants and build two
    dictionaries -- straight and reverse mappings. E.g.:

    ## neighbor attributes
    NDA_UNSPEC = 0
    NDA_DST = 1
    NDA_LLADDR = 2
    NDA_CACHEINFO = 3
    NDA_PROBES = 4
    (NDA_NAMES, NDA_VALUES) = map_namespace('NDA', globals())

    Will lead to::

        NDA_NAMES = {'NDA_UNSPEC': 0,
                     ...
                     'NDA_PROBES': 4}
        NDA_VALUES = {0: 'NDA_UNSPEC',
                      ...
                      4: 'NDA_PROBES'}

    The `normalize` parameter can be:

        - None — no name transformation will be done
        - True — cut the prefix and `lower()` the rest
        - lambda x: … — apply the function to every name
    '''
    transform: Callable[[str], str]

    if normalize is None:
        transform = _no_change
    elif normalize is True:
        transform = partial(_default_normalize, prefix=prefix)
    elif isinstance(normalize, types.FunctionType):
        transform = normalize
    else:
        raise ValueError("Invalid value for `normalize` parameter")

    by_name = {transform(i): ns[i] for i in ns.keys() if i.startswith(prefix)}
    by_value = {ns[i]: transform(i) for i in ns.keys() if i.startswith(prefix)}
    return (by_name, by_value)


def getbroadcast(
    addr: str, mask: int, family: socket.AddressFamily = socket.AF_INET
) -> str:
    # 1. convert addr to int
    i = socket.inet_pton(family, addr)
    if family == socket.AF_INET:
        i_unpacked = struct.unpack('>I', i)[0]
        a = 0xFFFFFFFF
        length = 32
    elif family == socket.AF_INET6:
        i_unpacked = struct.unpack('>QQ', i)
        i_unpacked = i_unpacked[0] << 64 | i_unpacked[1]
        a = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
        length = 128
    else:
        raise NotImplementedError('family not supported')
    # 2. calculate mask
    m = (a << length - mask) & a
    # 3. calculate default broadcast
    n = (i_unpacked & m) | a >> mask
    # 4. convert it back to the normal address form
    if family == socket.AF_INET:
        n = struct.pack('>I', n)
    else:
        n = struct.pack('>QQ', n >> 64, n & (a >> 64))
    return socket.inet_ntop(family, n)


def dqn2int(mask: str, family: socket.AddressFamily = socket.AF_INET) -> int:
    '''
    IPv4 dotted quad notation to int mask conversion
    '''
    ret = 0
    binary = socket.inet_pton(family, mask)
    for offset in range(len(binary) // 4):
        ret += bin(
            struct.unpack('I', binary[offset * 4 : offset * 4 + 4])[0]
        ).count('1')
    return ret


def get_address_family(address: str) -> socket.AddressFamily:
    if address.find(':') > -1:
        return socket.AF_INET6
    else:
        return socket.AF_INET


def hexdump(payload: bytes, length: int = 0) -> str:
    '''
    Represent byte string as hex -- for debug purposes
    '''
    return ':'.join('{0:02x}'.format(c) for c in payload[:length] or payload)


def load_dump(
    f: Union[str, io.StringIO], meta: Optional[dict[str, str]] = None
) -> Union[bytes, str]:
    '''
    Load a packet dump from an open file-like object or a string.

    Supported dump formats:

    * strace hex dump (\\x00\\x00...)
    * pyroute2 hex dump (00:00:...)

    Simple markup is also supported. Any data from # or ;
    till the end of the string is a comment and ignored.
    Any data after . till EOF is ignored as well.

    With #! starts an optional code block. All the data
    in the code block will be read and returned via
    metadata dictionary.
    '''
    data = ''
    code = None
    meta_data = None
    meta_label = None
    io_obj: Union[io.StringIO, str]
    if isinstance(f, str):
        io_obj = io.StringIO()
        io_obj.write(f)
        io_obj.seek(0)
    else:
        io_obj = f

    for a in io_obj.readlines():
        if code is not None:
            code += a
            continue

        if meta_data is not None:
            meta_data += a
            continue

        offset = 0
        length = len(a)
        while offset < length:
            if a[offset] in (' ', '\t', '\n'):
                offset += 1
            elif a[offset] == '#':
                if a[offset : offset + 2] == '#!':
                    # read the code block until EOF
                    code = ''
                elif a[offset : offset + 2] == '#:':
                    # read data block until EOF
                    meta_label = a.split(':')[1].strip()
                    meta_data = ''
                break
            elif a[offset] == '.':
                return data
            elif a[offset] == '\\':
                # strace hex format
                data += chr(int(a[offset + 2 : offset + 4], 16))
                offset += 4
            else:
                # pyroute2 hex format
                data += chr(int(a[offset : offset + 2], 16))
                offset += 3

    if isinstance(meta, dict):
        if code is not None:
            meta['code'] = code
        if meta_data is not None and meta_label is not None:
            meta[meta_label] = meta_data

    return bytes(data, 'iso8859-1')


class AddrPool(object):
    '''
    Address pool
    '''

    cell = 0xFFFFFFFFFFFFFFFF

    def __init__(
        self,
        minaddr: int = 0xF,
        maxaddr: int = 0xFFFFFF,
        reverse: bool = False,
        release: bool = False,
    ):
        self.cell_size = 0  # in bits
        mx = self.cell
        self.reverse = reverse
        self.release = release
        self.ban: list[dict[str, int]] = []
        while mx:
            mx >>= 8
            self.cell_size += 1
        self.cell_size *= 8
        # calculate, how many ints we need to bitmap all addresses
        self.cells = int((maxaddr - minaddr) / self.cell_size + 1)
        # initial array
        self.addr_map = [self.cell]
        self.minaddr = minaddr
        self.maxaddr = maxaddr
        self.lock = threading.RLock()

    def alloc(self) -> int:
        with self.lock:
            # gc self.ban:
            for item in tuple(self.ban):
                if item['counter'] == 0:
                    self.free(item['addr'])
                    self.ban.remove(item)
                else:
                    item['counter'] -= 1

            # iterate through addr_map
            base = 0
            for cell in self.addr_map:
                if cell:
                    # not allocated addr
                    bit = 0
                    while True:
                        if (1 << bit) & self.addr_map[base]:
                            self.addr_map[base] ^= 1 << bit
                            break
                        bit += 1
                    ret = base * self.cell_size + bit

                    if self.reverse:
                        ret = self.maxaddr - ret
                    else:
                        ret = ret + self.minaddr

                    if self.minaddr <= ret <= self.maxaddr:
                        if self.release:
                            self.free(ret, ban=self.release)
                        return ret
                    else:
                        self.free(ret)
                        raise KeyError('no free address available')

                base += 1
            # no free address available
            if len(self.addr_map) < self.cells:
                # create new cell to allocate address from
                self.addr_map.append(self.cell)
                return self.alloc()
            else:
                raise KeyError('no free address available')

    def locate(self, addr: int) -> tuple[int, int, bool]:
        if self.reverse:
            addr = self.maxaddr - addr
        else:
            addr -= self.minaddr
        base = addr // self.cell_size
        bit = addr % self.cell_size
        try:
            is_allocated = not self.addr_map[base] & (1 << bit)
        except IndexError:
            is_allocated = False
        return (base, bit, is_allocated)

    def free(self, addr: int, ban: int = 0):
        with self.lock:
            if ban != 0:
                self.ban.append({'addr': addr, 'counter': ban})
            else:
                base, bit, _ = self.locate(addr)
                if len(self.addr_map) <= base:
                    raise KeyError('address is not allocated')
                if self.addr_map[base] & (1 << bit):
                    raise KeyError('address is not allocated')
                self.addr_map[base] ^= 1 << bit


def fnv1(data: bytes) -> int:
    '''
    FNV1 -- 32bit hash, python3 version

    Returns: 32bit int hash

    See: http://www.isthe.com/chongo/tech/comp/fnv/index.html
    '''
    hval = 0x811C9DC5
    for i in range(len(data)):
        hval *= 0x01000193
        hval ^= data[i]
    return hval & 0xFFFFFFFF


def uuid32() -> int:
    '''
    Return 32bit UUID, based on the current time and pid.

    The uuid is guaranteed to be unique within one process.
    '''
    global _uuid32
    global _uuid32_lock

    with _uuid32_lock:
        candidate = _uuid32
        while candidate == _uuid32:
            candidate = fnv1(
                struct.pack('QQ', int(time.time() * 1000000), os.getpid())
            )
        _uuid32 = candidate
        return candidate


def uifname() -> str:
    '''
    Return a unique interface name based on a prime function
    '''
    return 'pr%x' % uuid32()


F = TypeVar("F", bound=Callable[..., Any])


def map_exception(
    match: Callable[[Exception], bool], subst: Callable[[Exception], Exception]
) -> Callable[[F], F]:
    '''
    Decorator to map exception types
    '''

    def wrapper(f: F) -> F:
        def decorated(*argv: Any, **kwarg: Any) -> Any:
            try:
                f(*argv, **kwarg)
            except Exception as e:
                if match(e):
                    raise subst(e)
                raise

        return decorated  # type: ignore

    return wrapper


def map_enoent(f: F) -> F:
    '''
    Shortcut to map OSError(2) -> OSError(95)
    '''
    return map_exception(
        lambda x: (isinstance(x, OSError) and x.errno == errno.ENOENT),
        lambda x: OSError(errno.EOPNOTSUPP, 'Operation not supported'),
    )(f)


T = TypeVar('T', bound=type)


def metaclass(mc: T) -> Callable[[T], T]:
    def wrapped(cls: T) -> T:
        nvars = {}
        skip = ['__dict__', '__weakref__']
        slots = cls.__dict__.get('__slots__')
        if not isinstance(slots, (list, tuple)):
            slots = [slots]
        for k in slots:
            skip.append(k)
        for k, v in cls.__dict__.items():
            if k not in skip:
                nvars[k] = v
        return mc(cls.__name__, cls.__bases__, nvars)

    return wrapped


def msg_done(msg) -> bytes:
    newmsg = struct.pack('IHH', 40, 2, 0)
    newmsg += msg.data[8:16]
    newmsg += struct.pack('I', 0)
    # nlmsgerr struct alignment
    newmsg += b'\0' * 20
    return newmsg


def get_time() -> float:
    '''
    Return seconds since arbitrary start point.

    It cannot go backward. It includes any time that the system is suspended.
    You should use this instead of time.time() to measure time between two
    execution points, like a timeout.
    '''
    return time.monotonic()



from socket import AF_INET

from .common import NLAKeyTransform


class ProbeFieldFilter(NLAKeyTransform):
    _nla_prefix = 'PROBE_'

    def finalize(self, context):
        if 'family' not in context:
            context['family'] = AF_INET
        if 'proto' not in context:
            context['proto'] = 1
        if 'port' not in context:
            context['port'] = 0
        if 'dst_len' not in context:
            context['dst_len'] = 32
        if 'kind' not in context:
            context['kind'] = 'ping'
        if 'num' not in context:
            context['num'] = 1
        if 'timeout' not in context:
            context['timeout'] = 1



from pyroute2.netlink import nla, nlmsg_atoms


class geneve(nla):
    nla_map = (
        ('IFLA_GENEVE_UNSPEC', 'none'),
        ('IFLA_GENEVE_ID', 'uint32'),
        ('IFLA_GENEVE_REMOTE', 'ip4addr'),
        ('IFLA_GENEVE_TTL', 'uint8'),
        ('IFLA_GENEVE_TOS', 'uint8'),
        ('IFLA_GENEVE_PORT', 'be16'),
        ('IFLA_GENEVE_COLLECT_METADATA', 'flag'),
        ('IFLA_GENEVE_REMOTE6', 'ip6addr'),
        ('IFLA_GENEVE_UDP_CSUM', 'uint8'),
        ('IFLA_GENEVE_UDP_ZERO_CSUM6_TX', 'uint8'),
        ('IFLA_GENEVE_UDP_ZERO_CSUM6_RX', 'uint8'),
        ('IFLA_GENEVE_LABEL', 'be32'),
        ('IFLA_GENEVE_TTL_INHERIT', 'uint8'),
        ('IFLA_GENEVE_DF', 'df'),
    )

    class df(nlmsg_atoms.uint16):
        value_map = {0: 'unset', 1: 'set', 2: 'inherit'}



from pyroute2 import IPRoute
from pyroute2 import NetlinkError as E1
from pyroute2.netlink import NetlinkError as E2
from pyroute2.netlink.exceptions import NetlinkError as E3


def test_exceptions():
    assert E1 == E2 == E3
    with IPRoute() as ipr:
        for e in (E1, E2, E3):
            try:
                ipr.get_links(-1)
            except e:
                pass



'''
Accessing objects
=================

NDB objects are grouped into "views":

    * interfaces
    * addresses
    * routes
    * neighbours
    * rules
    * netns
    * ...

Views are dictionary-like objects that accept strings or dict selectors::

    # access eth0
    ndb.interfaces["eth0"]

    # access eth0 in the netns test01
    ndb.sources.add(netns="test01")
    ndb.interfaces[{"target": "test01", "ifname": "eth0"}]

    # access a route to 10.4.0.0/24
    ndb.routes["10.4.0.0/24"]

    # same with a dict selector
    ndb.routes[{"dst": "10.4.0.0", "dst_len": 24}]

Objects cache
=============

NDB create objects on demand, it doesn't create thousands of route objects
for thousands of routes by default. The object is being created only when
accessed for the first time, and stays in the cache as long as it has any
not committed changes. To inspect cached objects, use views' `.cache`::

    >>> ndb.interfaces.cache.keys()
    [(('target', u'localhost'), ('tflags', 0), ('index', 1)),  # lo
     (('target', u'localhost'), ('tflags', 0), ('index', 5))]  # eth3

There is no asynchronous cache invalidation, the cache is being cleaned up
every time when an object is accessed.

API
===
'''

import asyncio
import errno
import gc
import json
import threading
import time
from collections import OrderedDict
from functools import partial

from pyroute2 import config
from pyroute2.common import basestring

##
# NDB stuff
from .objects import RSLV_DELETE
from .objects.address import Address
from .objects.interface import Interface, Vlan
from .objects.neighbour import FDBRecord, Neighbour
from .objects.netns import NetNS
from .objects.probe import Probe
from .objects.route import Route
from .objects.rule import Rule
from .report import Record, RecordSet
from .source import Source, SourceProxy


class TmpHandler:
    def __init__(self, ndb, event, handler):
        self.ndb = ndb
        self.event = event
        self.handler = handler

    def __enter__(self):
        self.ndb.task_manager.register_handler(
            self.ndb.schema.classes[self.event], self.handler
        )
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.ndb.task_manager.unregister_handler(
            self.ndb.schema.classes[self.event], self.handler
        )


class View(dict):
    '''
    The View() object returns RTNL objects on demand::

        ifobj1 = ndb.interfaces['eth0']
        ifobj2 = ndb.interfaces['eth0']
        # ifobj1 != ifobj2
    '''

    def __init__(self, ndb, table, chain=None):
        self.ndb = ndb
        self.log = ndb.log.channel('view.%s' % table)
        self.table = table
        self.event = table  # FIXME
        self.chain = chain
        self.cache = {}
        self.constraints = {}
        self.classes = OrderedDict()
        self.classes['interfaces'] = Interface
        self.classes['addresses'] = Address
        self.classes['neighbours'] = Neighbour
        self.classes['af_bridge_fdb'] = FDBRecord
        self.classes['routes'] = Route
        self.classes['rules'] = Rule
        self.classes['netns'] = NetNS
        self.classes['probes'] = Probe
        self.classes['af_bridge_vlans'] = Vlan

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    @property
    def default_target(self):
        if self.table == 'netns':
            return self.ndb.nsmanager
        else:
            return self.ndb.localhost

    @property
    def context(self):
        if self.chain is not None:
            return self.chain.context
        else:
            return {}

    def getmany(self, spec, table=None):
        return self.ndb.schema.get(table or self.table, spec)

    def getone(self, spec, table=None):
        for obj in self.getmany(spec, table):
            return obj

    def get(self, spec=None, table=None, **kwarg):
        spec = spec or kwarg
        try:
            return self.__getitem__(spec, table)
        except KeyError:
            return None

    def template(self, key, table=None):
        if self.chain:
            context = self.chain.context
        else:
            context = {}
        iclass = self.classes[table or self.table]
        spec = iclass.new_spec(key, context, self.default_target)
        return iclass(self, spec, load=False, master=self.chain)

    def create(self, *argspec, **kwspec):
        iclass = self.classes[self.table]
        if self.chain:
            context = self.chain.context
        else:
            context = {}
        spec = iclass.new_spec(
            kwspec or argspec[0], context, self.default_target
        )
        if self.chain:
            spec['ndb_chain'] = self.chain
        spec['create'] = True
        return self[spec]

    def ensure(self, *argspec, **kwspec):
        try:
            obj = self.locate(**kwspec)
        except KeyError:
            obj = self.create(**kwspec)
        for key, value in kwspec.items():
            obj[key] = value
        return obj

    def add(self, *argspec, **kwspec):
        self.log.warning(
            '''\n
        The name add() will be removed in future releases, use create()
        instead. If you believe that the idea to rename is wrong, please
        file your opinion to the project's bugtracker.

        The reason behind the rename is not to confuse interfaces.add() with
        bridge and bond port operations, that don't create any new interfaces
        but work on existing ones.
        '''
        )
        return self.create(*argspec, **kwspec)

    async def wait(self, **spec):
        ret = None
        timeout = spec.pop('timeout', -1)
        action = spec.pop('action', 'add')
        ctime = time.time()

        # install a limited events queue -- for a possible immediate reaction
        evq = asyncio.Queue(maxsize=100)

        async def handler(evq, target, event):
            # ignore the "queue full" exception
            #
            # if we miss some events here, nothing bad happens: we just
            # load them from the DB after a timeout, falling back to
            # the DB polling
            #
            # the most important here is not to allocate too much memory
            try:
                evq.put_nowait((target, event))
            except asyncio.queues.QueueFull:
                pass

        with TmpHandler(self.ndb, self.event, partial(handler, evq)):
            while True:
                ret = self.get(spec)
                if (ret and action == 'add') or (
                    ret is None and action == 'remove'
                ):
                    return ret
                try:
                    target, msg = await asyncio.wait_for(evq.get(), 1)
                except asyncio.TimeoutError:
                    pass
                if timeout > -1:
                    if ctime + timeout < time.time():
                        raise TimeoutError()

    def locate(self, spec=None, table=None, **kwarg):
        '''
        This method works like `__getitem__()`, but the important
        difference is that it uses only key fields to locate the
        object in the DB, ignoring all other keys.

        It is useful to locate objects that may change attributes
        during request, like an interface may come up/down, or an
        address may become primary/secondary, so plain
        `__getitem__()` will not match while the object still
        exists.
        '''
        if isinstance(spec, Record):
            spec = spec._as_dict()
        spec = spec or kwarg
        if not spec:
            raise TypeError('got an empty spec')

        table = table or self.table
        iclass = self.classes[table]
        spec = iclass.new_spec(spec)
        kspec = self.ndb.schema.compiled[table]['norm_idx']
        lookup_fallbacks = self.ndb.schema.compiled[table]['lookup_fallbacks']
        request = {}
        for name in kspec:
            name = iclass.nla2name(name)
            if name in spec:
                request[name] = spec[name]
            elif name in lookup_fallbacks:
                fallback = lookup_fallbacks[name]
                if fallback in spec:
                    request[fallback] = spec[fallback]

        if not request:
            raise KeyError('got an empty key')
        return self[request]

    def __getitem__(self, key, table=None):
        ret = self.template(key, table)

        # rtnl_object.key() returns a dictionary that can not
        # be used as a cache key. Create here a tuple from it.
        # The key order guaranteed by the dictionary.
        cache_key = tuple(ret.key.items())

        rtime = time.time()

        # Iterate all the cache to remove unused and clean
        # (without any started transaction) objects.
        for ckey in tuple(self.cache):
            # Skip the current cache_key to avoid extra
            # cache del/add records in the logs
            if ckey == cache_key:
                continue
            # 1. Remove only expired items
            # 2. The number of changed rtnl_object fields must
            #    be 0 which means that no transaction is started
            # 3. The number of referrers must be > 1, the first
            #    one is the cache itself        <- this op is expensive!
            if (
                rtime - self.cache[ckey].atime > config.cache_expire
                and self.cache[ckey].clean
                and gc.get_referrers(self.cache[ckey])
            ):
                self.log.debug('cache del %s' % (ckey,))
                self.cache.pop(ckey, None)

        if cache_key in self.cache:
            self.log.debug('cache hit %s' % (cache_key,))
            # Explicitly get rid of the created object
            del ret
            # The object from the cache has already
            # registered callbacks, simply return it
            ret = self.cache[cache_key]
            ret.atime = rtime
            return ret
        else:
            # Cache only existing objects
            if self.exists(key):
                ret.load_sql()
                self.log.debug('cache add %s' % (cache_key,))
                self.cache[cache_key] = ret

        ret.register()
        return ret

    def exists(self, key, table=None):
        '''
        Check if the specified object exists in the database::

            ndb.interfaces.exists('eth0')
            ndb.interfaces.exists({'ifname': 'eth0', 'target': 'localhost'})
            ndb.addresses.exists('127.0.0.1/8')
        '''
        if self.chain:
            context = self.chain.context
        else:
            context = {}

        iclass = self.classes[self.table]
        key = iclass.new_spec(key, context, self.default_target)

        iclass.resolve(
            view=self,
            spec=key,
            fields=iclass.resolve_fields,
            policy=RSLV_DELETE,
        )

        table = table or self.table
        schema = self.ndb.schema
        names = schema.compiled[self.table]['all_names']

        self.log.debug('check if the key %s exists in table %s' % (key, table))
        keys = []
        values = []
        for name, value in key.items():
            nla_name = iclass.name2nla(name)
            if nla_name in names:
                name = nla_name
            if value is not None and name in names:
                keys.append(f'f_{name} = ?')
                if isinstance(value, (dict, list, tuple, set)):
                    value = json.dumps(value)
                values.append(value)
        spec = schema.fetchone(
            'SELECT * FROM %s WHERE %s' % (self.table, ' AND '.join(keys)),
            values,
        )
        if spec is not None:
            self.log.debug('exists')
            return True
        else:
            self.log.debug('not exists')
            return False

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __delitem__(self, key):
        raise NotImplementedError()

    def __iter__(self):
        return self.keys()

    def __contains__(self, key):
        return key in self.dump()

    def keys(self):
        for record in self.dump():
            yield record

    def values(self):
        for key in self.keys():
            yield self[key]

    def items(self):
        for key in self.keys():
            yield (key, self[key])

    def count(self):
        return self.classes[self.table]._count(self)[0]

    def __len__(self):
        return self.count()

    def _keys(self, iclass):
        return ['target', 'tflags'] + self.ndb.schema.compiled[
            iclass.view or iclass.table
        ]['names']

    def _native(self, dump):
        fnames = next(dump)
        for record in dump:
            yield Record(fnames, record, self.classes[self.table])

    def dump(self):
        iclass = self.classes[self.table]
        return RecordSet(self._native(iclass.dump(self)))

    def summary(self):
        iclass = self.classes[self.table]
        return RecordSet(self._native(iclass.summary(self)))

    def __repr__(self):
        if self.chain and 'ifname' in self.chain:
            parent = f'{self.chain["ifname"]}/'
        else:
            parent = ''
        return f'''
NDB view for {parent}{self.table}
Number of objects: {self.count()}

to list objects use .summary() or .dump()
    -> RecordSet (generator)
        -> Record

key: Union[Record, dict, spec]
to get objects use ...[key] / .__getitem__(key)
    -> RTNL_Object
'''


class SourcesView(View):
    def __init__(self, ndb, table='sources'):
        super(SourcesView, self).__init__(ndb, table)
        self.classes[table] = Source
        self.cache = {}
        self.proxy = {}
        self.lock = threading.Lock()

    def async_add(self, **spec):
        spec = dict(Source.defaults(spec))
        self.cache[spec['target']] = Source(self.ndb, **spec).start()
        return self.cache[spec['target']]

    async def add(self, **spec):
        spec = dict(Source.defaults(spec))
        target = spec['target']
        if target in self:
            raise KeyError(f'source {target} exists')
        if spec.get('event') is None:
            sync = True
            spec['event'] = asyncio.Event()
        else:
            sync = False
        source = Source(self.ndb, **spec)
        self.cache[spec['target']] = source
        task = self.ndb.task_manager.create_task(source.receiver, obj=source)
        if sync:
            await task.event.wait()
            if task.exception:
                raise task.exception
        return self.cache[spec['target']]

    def remove(self, target, code=errno.ECONNRESET, sync=True):
        if target not in self:
            raise KeyError(f'source {target} does not exist')
        with self.lock:
            if target in self.cache:
                source = self.cache[target]
                task = self.ndb.task_manager.task_map[source.task_id]
                task.target_state = 'stopped'
                self.ndb.schema.flush(target)
                source.nl.close(code)
                return self.cache.pop(target)

    def keys(self):
        for key in self.cache:
            yield key

    def _keys(self, iclass):
        return ['target', 'kind']

    def wait(self, **spec):
        raise NotImplementedError()

    def _summary(self, *argv, **kwarg):
        return self._dump(*argv, **kwarg)

    def __getitem__(self, key, table=None):
        if isinstance(key, basestring):
            target = key
        elif isinstance(key, dict) and 'target' in key.keys():
            target = key['target']
        else:
            raise KeyError()

        if target in self.cache:
            return self.cache[target]
        elif target in self.proxy:
            return self.proxy[target]
        else:
            proxy = SourceProxy(self.ndb, target)
            self.proxy[target] = proxy
            return proxy



from pyroute2.netlink import nla


class vxlan(nla):
    prefix = 'IFLA_'
    nla_map = (
        ('IFLA_VXLAN_UNSPEC', 'none'),
        ('IFLA_VXLAN_ID', 'uint32'),
        ('IFLA_VXLAN_GROUP', 'ip4addr'),
        ('IFLA_VXLAN_LINK', 'uint32'),
        ('IFLA_VXLAN_LOCAL', 'ip4addr'),
        ('IFLA_VXLAN_TTL', 'uint8'),
        ('IFLA_VXLAN_TOS', 'uint8'),
        ('IFLA_VXLAN_LEARNING', 'uint8'),
        ('IFLA_VXLAN_AGEING', 'uint32'),
        ('IFLA_VXLAN_LIMIT', 'uint32'),
        ('IFLA_VXLAN_PORT_RANGE', 'port_range'),
        ('IFLA_VXLAN_PROXY', 'uint8'),
        ('IFLA_VXLAN_RSC', 'uint8'),
        ('IFLA_VXLAN_L2MISS', 'uint8'),
        ('IFLA_VXLAN_L3MISS', 'uint8'),
        ('IFLA_VXLAN_PORT', 'be16'),
        ('IFLA_VXLAN_GROUP6', 'ip6addr'),
        ('IFLA_VXLAN_LOCAL6', 'ip6addr'),
        ('IFLA_VXLAN_UDP_CSUM', 'uint8'),
        ('IFLA_VXLAN_UDP_ZERO_CSUM6_TX', 'uint8'),
        ('IFLA_VXLAN_UDP_ZERO_CSUM6_RX', 'uint8'),
        ('IFLA_VXLAN_REMCSUM_TX', 'uint8'),
        ('IFLA_VXLAN_REMCSUM_RX', 'uint8'),
        ('IFLA_VXLAN_GBP', 'flag'),
        ('IFLA_VXLAN_REMCSUM_NOPARTIAL', 'flag'),
        ('IFLA_VXLAN_COLLECT_METADATA', 'uint8'),
        ('IFLA_VXLAN_LABEL', 'uint32'),
        ('IFLA_VXLAN_GPE', 'flag'),
        ('IFLA_VXLAN_TTL_INHERIT', 'flag'),
        ('IFLA_VXLAN_DF', 'uint8'),
    )

    class port_range(nla):
        fields = (('low', '>H'), ('high', '>H'))



import io
import json
import shlex
import struct
from collections import namedtuple
from importlib import import_module

from pyroute2.common import hexdump, load_dump

PcapMetaData = namedtuple(
    "pCAPMetaData",
    (
        "magic_number",
        "version_major",
        "version_minor",
        "thiszone",
        "sigfigs",
        "snaplen",
        "network",
    ),
)
PcapPacketHeader = namedtuple(
    "PcapPacketHeader",
    ("ts_sec", "ts_usec", "incl_len", "orig_len", "header_len"),
)
PcapLLHeader = namedtuple(
    "PcapLLHeader",
    ("pad0", "addr_type", "pad1", "pad2", "pad3", "family", "header_len"),
)


class Message:

    def __init__(self, packet_header, ll_header, met, key, data):
        self.packet_header = packet_header
        self.ll_header = ll_header
        self.cls = None
        self.met = met
        self.key = key
        self.kl = struct.calcsize(self.key)
        self.data = data
        self.exception = None
        self.msg = None

    def get_message_class(self):
        if hasattr(self.met, 'msg_map'):
            (msg_type,) = struct.unpack(self.key, self.data[4 : 4 + self.kl])
            return self.met.msg_map[msg_type]
        return self.met

    def decode(self):
        try:
            self.cls = self.get_message_class()
            self.msg = self.cls(self.data)
            self.msg.decode()
            self.msg = self.msg.dump()
        except Exception as e:
            self.exception = repr(e)
            self.msg = hexdump(self.data)

    def dump(self):
        return {
            "pcap header": repr(self.packet_header),
            "link layer header": repr(self.ll_header),
            "message class": repr(self.cls),
            "exception": self.exception,
            "data": self.msg,
        }

    def __repr__(self):
        return json.dumps(self.dump(), indent=4)


class MatchOps:
    '''
    Functions to match netlink messages.

    The matcher object maintains a stack, where every function
    leaves True or False. A message matches only when the stack
    contains True.

    Some functions take arguments from the command line, other
    like `AND` and `OR` work with the stack.
    '''

    @staticmethod
    def AND():
        '''
        Consumes values left on the stack by functions to the
        left and to the right in the expression, and leaves
        the result of AND operation::

            func_a{...} AND func_b{...}
        '''

        def f(packet_header, ll_header, raw, data_offset, stack):
            v1 = stack.pop()
            v2 = stack.pop()
            return v1 and v2

        return f

    @staticmethod
    def OR():
        '''
        Consumes values left on the stack by functions to the
        left and to the right in the expression, and leaves
        the result of OR operation::

            func_a{...} OR func_b{...}
        '''

        def f(packet_header, ll_header, raw, data_offset, stack):
            v1 = stack.pop()
            v2 = stack.pop()
            return v1 or v2

        return f

    @staticmethod
    def ll_header(family):
        '''
        Match link layer header fields. As for now only netlink
        family is supported, see `pyroute2.netlink` for netlink
        families (`NETLINK_.*`) constants::

            # match generic netlink messages
            ll_header{family=16}

            # match RTNL messages
            ll_header{family=0}
        '''
        if not isinstance(family, int) or family < 0 or family > 0xFFFF:
            raise TypeError('family must be unsigned short integer')

        def f(packet_header, ll_header, raw, data_offset, stack):
            if ll_header is None:
                return False
            return ll_header.family == family

        return f

    @staticmethod
    def data(fmt, offset, value):
        '''
        Match a voluntary data in the message. Use `struct` notation
        for the format, integers for offset and value::

            # match four bytes with offset 4 bytes and value 16,
            # or 10:00:00:00 in hex:

            data{fmt='I', offset=4, value=16}

            # match one byte with offset 16 and value 1, or 01 in hex

            data{fmt='B', offset=16, value=1}

        More examples::

            # match:
            #   * generic netlink protocol, 16
            #   * message type 37 -- IPVS protocol for this session
            #   * message command 1 -- IPVS_CMD_NEW_SERVICE
            ll_header{family=16}
                AND data{fmt='H', offset=4, value=37}
                AND data{fmt='B', offset=16, value=1}
        '''
        if not isinstance(fmt, str):
            raise TypeError('format must be string')
        if not isinstance(offset, int) or not isinstance(value, int):
            raise TypeError('offset and value must be integers')

        def f(packet_header, ll_header, raw, data_offset, stack):
            o = data_offset + offset
            s = struct.calcsize(fmt)
            return struct.unpack(fmt, raw[o : o + s])[0] == value

        return f


class Matcher:
    def __init__(self, script):
        self.parsed = []
        self.filters = []
        self.script = script
        self.shlex = shlex.shlex(instream=io.StringIO(script))
        self.shlex.wordchars += '-~'
        postpone = None
        while True:
            token = self.get_token(ignore=',')
            if token == '':
                break
            method = getattr(MatchOps, token)
            if token in ('AND', 'OR'):
                postpone = method
                continue
            kwarg = {}
            token = self.get_token(expect='{')
            while True:
                token = self.get_token(ignore=',')
                if token in ('}', ''):
                    break
                self.get_token(expect='=')
                value = self.get_token()
                if value[0] in ['"', "'"]:
                    # string
                    value = value[1:-1]
                else:
                    # int
                    value = int(value)
                kwarg[token] = value
            self.filters.append(method(**kwarg))
            if postpone is not None:
                self.filters.append(postpone())
                postpone = None

    def get_token(self, expect=None, ignore=None):
        token = self.shlex.get_token()
        self.parsed.append(token)
        if expect is not None and token != expect:
            raise SyntaxError(f"expected {expect}: {' '.join(self.parsed)} <-")
        if ignore is not None and token in ignore:
            token = self.shlex.get_token()
            self.parsed.append(token)
        return token

    def match(self, packet_header, ll_header, data, offset):
        stack = []
        for method in self.filters:
            stack.append(method(packet_header, ll_header, data, offset, stack))
        return all(stack)


class LoaderHex:

    def __init__(self, data, cls, key, data_offset, script):
        with open(data, 'r') as f:
            self.raw = load_dump(f)
        self.cls = cls
        self.key = key
        self.offset = 0
        self.matcher = Matcher(script)

    @property
    def data(self):
        while self.offset < len(self.raw):
            msg = Message(
                None, None, self.cls, self.key, self.raw[self.offset :]
            )
            msg.decode()
            if self.matcher.match(None, None, self.raw, self.offset):
                yield msg
            self.offset += msg.msg['header']['length']


class LoaderPcap:

    def __init__(self, data, cls, key, data_offset, script):
        with open(data, 'rb') as f:
            self.raw = f.read()
        self.metadata = PcapMetaData(*struct.unpack("IHHiIII", self.raw[:24]))
        self.offset = 24
        self.key = key
        self.data_offset = data_offset
        self.cls = cls
        self.matcher = Matcher(script)

    def decode_packet_header(self, data, offset):
        return PcapPacketHeader(
            *struct.unpack("IIII", data[offset : offset + 16]) + (16,)
        )

    def decode_ll_header(self, data, offset):
        return PcapLLHeader(
            *struct.unpack(">HHIIHH", data[offset : offset + 16]) + (16,)
        )

    @property
    def data(self):
        while self.offset < len(self.raw):
            packet_header = self.decode_packet_header(self.raw, self.offset)
            self.offset += packet_header.header_len
            ll_header = self.decode_ll_header(self.raw, self.offset)
            self.offset += ll_header.header_len
            length = packet_header.incl_len - ll_header.header_len
            off_length = length - self.data_offset
            if self.matcher.match(
                packet_header, ll_header, self.raw, self.offset
            ):
                offset = self.offset + self.data_offset
                msg = Message(
                    packet_header,
                    ll_header,
                    self.cls,
                    self.key,
                    self.raw[offset : offset + off_length],
                )
                msg.decode()
                yield msg
            self.offset += length


def get_loader(args):
    if args.cls:
        cls = args.cls.replace('/', '.').split('.')
        module_name = '.'.join(cls[:-1])
        cls_name = cls[-1]
        module = import_module(module_name)
        cls = getattr(module, cls_name)

    if args.format == 'pcap':
        return LoaderPcap(args.data, cls, args.key, args.offset, args.match)
    elif args.format == 'hex':
        return LoaderHex(args.data, cls, args.key, args.offset, args.match)
    else:
        raise ValueError('data format not supported')



from collections import namedtuple

# from ethtool/ethtool-copy.h of ethtool repo

DUPLEX_HALF = 0x0
DUPLEX_FULL = 0x1
DUPLEX_UNKNOWN = 0xFF

LINK_DUPLEX_NAMES = {
    DUPLEX_HALF: "Half",
    DUPLEX_FULL: "Full",
    DUPLEX_UNKNOWN: "Unknown",
}

# Which connector port.
PORT_TP = 0x00
PORT_AUI = 0x01
PORT_MII = 0x02
PORT_FIBRE = 0x03
PORT_BNC = 0x04
PORT_DA = 0x05
PORT_NONE = 0xEF
PORT_OTHER = 0xFF
LINK_PORT_NAMES = {
    PORT_TP: "Twisted Pair",
    PORT_AUI: "AUI",
    PORT_MII: "MII",
    PORT_FIBRE: "FIBRE",
    PORT_BNC: "BNC",
    PORT_DA: "Direct Attach Copper",
    PORT_NONE: "NONE",
    PORT_OTHER: "Other",
}

# Which transceiver to use.
XCVR_INTERNAL = 0x00  # PHY and MAC are in the same package
XCVR_EXTERNAL = 0x01  # PHY and MAC are in different packages
LINK_TRANSCEIVER_NAMES = {XCVR_INTERNAL: "Internal", XCVR_EXTERNAL: "External"}

# Enable or disable autonegotiation.
AUTONEG_DISABLE = 0x00
AUTONEG_ENABLE = 0x01
LINK_AUTONEG_NAMES = {AUTONEG_DISABLE: "off", AUTONEG_ENABLE: "on"}

# MDI or MDI-X status/control - if MDI/MDI_X/AUTO is set then
# the driver is required to renegotiate link
ETH_TP_MDI_INVALID = 0x00  # status: unknown; control: unsupported
ETH_TP_MDI = 0x01  # status: MDI; control: force MDI
ETH_TP_MDI_X = 0x02  # status: MDI-X; control: force MDI-X
ETH_TP_MDI_AUTO = 0x03  # control: auto-select
LINK_TP_MDI_NAMES = {
    ETH_TP_MDI: "off",
    ETH_TP_MDI_X: "on",
    ETH_TP_MDI_AUTO: "auto",
}


LMBTypePort = 0
LMBTypeMode = 1
LMBTypeOther = -1
LinkModeBit = namedtuple('LinkModeBit', ('bit_index', 'name', 'type'))
LinkModeBits = (
    LinkModeBit(bit_index=0, name='10baseT/Half', type=LMBTypeMode),
    LinkModeBit(bit_index=1, name='10baseT/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=2, name='100baseT/Half', type=LMBTypeMode),
    LinkModeBit(bit_index=3, name='100baseT/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=4, name='1000baseT/Half', type=LMBTypeMode),
    LinkModeBit(bit_index=5, name='1000baseT/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=6, name='Autoneg', type=LMBTypeOther),
    LinkModeBit(bit_index=7, name='TP', type=LMBTypePort),
    LinkModeBit(bit_index=8, name='AUI', type=LMBTypePort),
    LinkModeBit(bit_index=9, name='MII', type=LMBTypePort),
    LinkModeBit(bit_index=10, name='FIBRE', type=LMBTypePort),
    LinkModeBit(bit_index=11, name='BNC', type=LMBTypePort),
    LinkModeBit(bit_index=12, name='10000baseT/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=13, name='Pause', type=LMBTypeOther),
    LinkModeBit(bit_index=14, name='Asym_Pause', type=LMBTypeOther),
    LinkModeBit(bit_index=15, name='2500baseX/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=16, name='Backplane', type=LMBTypeOther),
    LinkModeBit(bit_index=17, name='1000baseKX/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=18, name='10000baseKX4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=19, name='10000baseKR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=20, name='10000baseR_FEC', type=LMBTypeMode),
    LinkModeBit(bit_index=21, name='20000baseMLD2/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=22, name='20000baseKR2/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=23, name='40000baseKR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=24, name='40000baseCR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=25, name='40000baseSR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=26, name='40000baseLR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=27, name='56000baseKR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=28, name='56000baseCR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=29, name='56000baseSR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=30, name='56000baseLR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=31, name='25000baseCR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=32, name='25000baseKR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=33, name='25000baseSR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=34, name='50000baseCR2/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=35, name='50000baseKR2/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=36, name='100000baseKR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=37, name='100000baseSR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=38, name='100000baseCR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=39, name='100000baseLR4_ER4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=40, name='50000baseSR2/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=41, name='1000baseX/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=42, name='10000baseCR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=43, name='10000baseSR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=44, name='10000baseLR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=45, name='10000baseLRM/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=46, name='10000baseER/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=47, name='2500baseT/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=48, name='5000baseT/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=49, name='FEC_NONE', type=LMBTypeOther),
    LinkModeBit(bit_index=50, name='FEC_RS', type=LMBTypeOther),
    LinkModeBit(bit_index=51, name='FEC_BASER', type=LMBTypeOther),
    LinkModeBit(bit_index=52, name='50000baseKR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=53, name='50000baseSR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=54, name='50000baseCR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=55, name='50000baseLR_ER_FR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=56, name='50000baseDR/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=57, name='100000baseKR2/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=58, name='100000baseSR2/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=59, name='100000baseCR2/Full', type=LMBTypeMode),
    LinkModeBit(
        bit_index=60, name='100000baseLR2_ER2_FR2/Full', type=LMBTypeMode
    ),
    LinkModeBit(bit_index=61, name='100000baseDR2/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=62, name='200000baseKR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=63, name='200000baseSR4/Full', type=LMBTypeMode),
    LinkModeBit(
        bit_index=64, name='200000baseLR4_ER4_FR4/Full', type=LMBTypeMode
    ),
    LinkModeBit(bit_index=65, name='200000baseDR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=66, name='200000baseCR4/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=67, name='100baseT1/Full', type=LMBTypeMode),
    LinkModeBit(bit_index=68, name='1000baseT1/Full', type=LMBTypeMode),
)
LinkModeBits_by_index = {bit.bit_index: bit for bit in LinkModeBits}



'''

One object
----------

All the changes done using one object are applied in the order defined by
the corresponding object class.

.. code-block:: python

    eth0 = ndb.interfaces["eth0"]
    eth0.add_ip(address="10.0.0.1", prefixlen=24)
    eth0.set(state="up")
    eth0.set(mtu=1400)
    eth0.commit()

In the example above first the interface attributes like state, mtu, ifname
etc. will be applied, and only then IP addresses, bridge ports and like that,
regardless the order they are referenced before the `commit()` call.

The order is ok for most of cases. But if not, one can control it by calling
`commit()` in the required places, breaking one transaction into several
sequential transactions.

And since RTNL object methods return the object itself, it is possible to
write chains with multiple `commit()`:

.. code-block:: python

    (
        ndb.interfaces
        .create(ifname="test", kind="dummy")
        .add_ip(address="10.0.0.1", prefixlen=24)
        .commit()
        .set(state="up")
        .commit()
    )

Here the order is forced by explicit commits.

Multiple objects
----------------

An important functionality of NDB are rollbacks. And there is a way to batch
changes on multiple objects so one failure will trigger rollback of all the
changes on all the objects.

.. code-block:: python

    ctx = ndb.begin()
    ctx.push(
        # first set up a bridge
        (
            ndb.interfaces
            .create(ifname="br0", kind="bridge")
            .add_port("eth0")
            .add_port("eth1")
            .set(state="up")
            .add_ip("10.0.0.2/24")
        ),
        # and only then create a route
        (
            ndb.routes
            .create(
                dst="192.168.0.0",
                dst_len=24,
                gateway="10.0.0.1"
            )
        )
    )
    ctx.commit()  # if something goes wrong, the whole batch will be reverted


Ping a remote host
------------------

The simplest usecase for external checks is to test if a remote IP is still
reachable after the changes are applied:

.. code-block:: python

    from pyroute2.ndb.transaction import PingAddress

    ctx = ndb.begin()
    ctx.push(
        ndb.routes.create(dst="10.0.0.0", dst_len=24, gateway="172.16.0.1"),
        PingAddress("10.0.0.1")
    )
    ctx.commit()  # the route will be removed if ping fails

Or on the contrary, don't run transaction if a remote IP is reachable:

.. code-block:: python

    from pyroute2.ndb.transaction import Not, PingAddress

    ctx = ndb.begin()
    ctx.push(
        Not(PingAddress("10.0.0.1")),
        ndb.routes.create(dst="10.0.0.0", dst_len=24, gateway="172.16.0.1")
    )
    try:
        ctx.commit()
    except CheckProcessException:
        pass

In this example, the route will be added only if `10.0.0.1` is not reachable.

The default ping timeout is set to 1, but it is possible to customize it:

.. code-block:: python

    PingAddress("10.0.0.1", timeout=10)

Check an external processes
---------------------------

A more generic type of check is CheckProcess:

.. code-block:: python

    from pyroute2.ndb.transaction import CheckProcess

    with ndb.begin() as ctx:
        ctx.push(ndb.routes.create(
            dst="10.0.0.0",
            dst_len=24,
            gateway="172.16.0.1"
        ))
        ctx.push(CheckProcess('/path/to/script.sh'))
        #
        # --> <-- the route will be removed if the script fails

`CheckProcess` is `subprocess.Popen` based, is not a shell call, thus no pipes
or other shell syntax are allowed.

`CheckProcess` also accepts `timeout` argument:

.. code-block:: python

    CheckProcess('/path/to/script.sh', timeout=10).commit()

If the subprocess doens't finish within the timeout, it will be terminated
with SIGTERM. SIGKILL is not used.

Logging and debug
-----------------

`CheckProcess` and `PingAddress` accept log as an argument:

.. code-block:: python

    PingAddress("10.0.0.1", log=ndb.log.channel("util")).commit()
    CheckProcess("/path/to/script.sh", log=ndb.log.channel("util")).commit()

The check objects are thread safe and reusable, it is possible to run
`commit()` on them multiple times. The subprocess' stdout and stderr will
be both logged and saved:

.. code-block:: python

    check = CheckProcess("/path/to/script.sh")
    while True:
        check.commit()  # periodic check, the loop breaks on failure
        print(f'stdout: {check.out}')
        print(f'stderr: {check.err}')
        print(f'return code: {check.return_code}')
        time.sleep(10)

Check negation
--------------

It is possible to negate the check for `CheckProcess` and child classes

.. code-block:: python

    from pyroute2.ndb.transaction import Not, CheckProcess

    check = Not(CheckProcess('/path/to/script.sh'))
    check.commit()

API
---

'''

import logging
import shlex
import shutil
import subprocess
import threading

global_log = logging.getLogger(__name__)


class CheckProcessException(Exception):
    pass


class CheckProcess:
    '''
    Run an external process on `commit()` and raise `CheckProcessException`
    if the return code is not 0.

    Objects of this class are thread safe and reusable.
    '''

    def __init__(self, command, log=None, timeout=None):
        if not isinstance(command, str):
            raise TypeError('command must be a non empty string')
        if not len(command) > 0:
            raise TypeError('command must be a non empty string')
        self.log = log or global_log
        self.command = command
        self.args = shlex.split(command)
        self.timeout = timeout
        self.return_code = None
        self.out = None
        self.err = None
        self.lock = threading.Lock()

    def commit(self):
        with self.lock:
            self.args[0] = shutil.which(self.args[0])
            if self.args[0] is None:
                raise FileNotFoundError()
            process = subprocess.Popen(
                self.args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            try:
                self.log.debug(f'process check {self.args}')
                self.out, self.err = process.communicate(timeout=self.timeout)
                self.log.debug(f'process output: {self.out}')
                self.log.debug(f'process stderr: {self.err}')
            except subprocess.TimeoutExpired:
                self.log.debug('process timeout expired')
                process.terminate()
                process.stdout.close()
                process.stderr.close()
            finally:
                self.return_code = process.wait()
            if self.return_code != 0:
                raise CheckProcessException('CheckProcess failed')

    def rollback(self):
        pass

    def __repr__(self):
        return f'[{self.command}]'


class PingAddress(CheckProcess):
    def __init__(self, address, log=None, timeout=1):
        super(PingAddress, self).__init__(
            f'ping -c 1 -W {timeout} {address}', log=log
        )


class Not:
    '''
    Negate the `CheckProcess` results. If `CheckProcess.commit()` succeeds,
    raise CheckProcessException, and vice versa, if `CheckProcess.commit()`
    fails, return success.
    '''

    def __init__(self, transaction):
        self.tx = transaction

    def commit(self):
        success = True
        try:
            self.tx.commit()
        except Exception:
            success = False
        if success:
            raise CheckProcessException(f'{self.tx} succeeded')

    def rollback(self):
        pass


class Transaction:
    '''
    `Transaction` class is an independent utility class. Being designed to
    be used with NDB object transactions, it may be used with any object
    implementing commit/rollback protocol, see `commit()` method.

    The class supports the context manager protocol and `Transaction` objects
    may be used in `with` statements:

    .. code-block:: python

        with Transaction() as tx:
            tx.push(obj0)  # enqueue objects
            tx.push(obj1)
            # --> <-- run commit() for every object in self.queue
            #
            # if any commit() fails, run rollback() for every
            # executed commit() in the reverse order

    NDB provides a utility method to create `Transaction` objects:

    .. code-block:: python

        with ndb.begin() as tx:
            tx.push(ndb.interfaces["eth0"].set(state="up"))
            tx.push(ndb.interfaces["eth1"].set(state="up"))
    '''

    def __init__(self, log=None):
        self.queue = []
        self.event = threading.Event()
        self.event.clear()
        self.log = global_log or log
        self.log.debug('begin transaction')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.commit()

    def push(self, *argv):
        '''
        Push objects to the transaction queue. One may use any number
        of positional arguments:

        .. code-block:: python

            tx.push(obj0)
            tx.push(obj0, obj1, obj2)
            tx.push(*[obj0, obj1, obj2])
        '''
        for obj in argv:
            self.log.debug('queue %s' % type(obj))
            self.queue.append(obj)
        return self

    def append(self, obj):
        '''
        Append one object to the queue.
        '''
        self.log.debug('queue %s' % type(obj))
        self.push(obj)
        return self

    def pop(self, index=-1):
        '''
        Pop an object from the queue. If an index is not specified,
        pop the last (the rightmost) object.
        '''
        self.log.debug('pop %s' % index)
        self.queue.pop(index)
        return self

    def insert(self, index, obj):
        '''
        Insert an object into the queue. The position index is required.
        '''
        self.log.debug('insert %i %s' % (index, type(obj)))
        self.queue.insert(index, obj)
        return self

    def cancel(self):
        '''
        Cancel the transaction and empty the queue.
        '''
        self.log.debug('cancel transaction')
        self.queue = []
        return self

    def wait(self, timeout=None):
        '''
        Wait until the transaction to be successfully committed.
        '''
        return self.event.wait(timeout)

    def done(self):
        '''
        Check if the done event is set.
        '''
        return self.event.is_set()

    def commit(self):
        '''
        Execute `commit()` for every queued object. If an execution
        fails, execute `rollback()` for every executed commit. All
        objects in the queue that follows the failed one will remain
        intact.

        Raises the original exception of the failed `commit()`. All
        the `rollback()` exceptions are ignored.
        '''
        self.log.debug('commit')
        rollbacks = []
        for obj in self.queue:
            rollbacks.append(obj)
            try:
                obj.commit()
            except Exception:
                for rb in reversed(rollbacks):
                    try:
                        rb.rollback()
                    except Exception as e:
                        self.log.warning('ignore rollback exception: %s' % e)
                raise
        self.event.set()
        return self



import errno
from collections.abc import AsyncGenerator, Generator
from typing import Generic, TypeVar

import pytest
import pytest_asyncio

from pyroute2 import netns
from pyroute2.common import uifname
from pyroute2.iproute.linux import AsyncIPRoute, IPRoute
from pyroute2.netlink.exceptions import NetlinkError
from pyroute2.netlink.rtnl import IFNAMSIZ
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg

T = TypeVar('T', AsyncIPRoute, IPRoute)


class TestInterface:
    '''Test interface spec.

    Provided by `test_link` fixture.

    Provides shortcuts to some important interface properties,
    like `TestInterface.index` or `TestInterface.netns`.
    '''

    def __init__(self, index: int, ifname: str, address: str, nsname: str):
        if index <= 1:
            raise TypeError('test interface index must be > 1')
        if not 1 < len(ifname) <= IFNAMSIZ:
            raise TypeError(
                'test interface ifname length must be from 2 to IFNAMSIZ'
            )
        self._index = index
        self._ifname = ifname
        self._address = address
        self._netns = nsname

    @property
    def index(self) -> int:
        '''Test interface index.

        Is always greater than 1, as index 1 has the the loopback interface.
        '''
        return self._index

    @property
    def ifname(self) -> str:
        '''Test interface ifname.

        The name length must be greater than 1 and less or equal IFNAMSIZ.
        '''
        return self._ifname

    @property
    def address(self) -> str:
        '''Test interface MAC address.

        In the form `xx:xx:xx:xx:xx:xx`.'''
        return self._address

    @property
    def netns(self) -> str:
        '''Test interface netns.

        A string name of the network namespace.'''
        return self._netns


class TestContext(Generic[T]):
    '''The test context.

    Provided by `async_context` and `sync_context` fixtures.

    Provides convenient shortcuts to RTNL API, the network namespace
    name and the test interface spec.
    '''

    __test__ = False  # prevent pytest from looking for tests here

    def __init__(self, ipr: T, test_link: TestInterface):
        self._ipr: T = ipr
        self._test_link = test_link

    @property
    def ipr(self) -> T:
        '''RTNL API.

        Return RTNL API instance, either IPRoute, or AsyncIPRoute.'''
        return self._ipr

    @property
    def test_link(self) -> TestInterface:
        '''Test interface spec.

        Return `TestInterface` object for the test interface.'''
        return self._test_link

    @property
    def netns(self) -> str:
        '''Network namespace.

        A string name of the network namespace.'''
        return self.ipr.status['netns']


class SetNSContext:
    '''A unique network namespace context.

    Provided by `setns_context` fixture.

    Sets a unique netns for the whole python process for the
    fixture scope, and returns from the nets on the scope exit.'''

    def __init__(self, nsname: str):
        self._netns = nsname

    def __enter__(self):
        netns.pushns(self.netns)

    def __exit__(self, *_):
        netns.popns()

    @property
    def netns(self) -> str:
        '''Network namespace.

        A string name of the network namespace.'''
        return self._netns


@pytest.fixture
def nsname() -> Generator[str]:
    '''Network namespace.

    * **Name**: nsname
    * **Scope**: function

    Create a unique network namespace and yield its name. Remove
    the netns on cleanup.

    It's safe to create and modify interfaces, addresses, routes etc.
    in the test network namespace, as it is disconnected from the main
    system, and the test cleanup will remove the namespace with all
    its content.

    Example usage:

    .. testcode:: nsname

        def test_list_interfaces(nsname):
            subprocess.Popen(
                ['ip', 'netns', 'exec', nsname, 'ip', 'link'],
                stdout=subprocess.PIPE,
            )
            # ...
    '''
    nsname = uifname()
    netns.create(nsname)
    with IPRoute(netns=nsname) as ipr:
        ipr.link('set', index=1, state='up')
        ipr.poll(ipr.addr, 'dump', address='127.0.0.1', timeout=5)
    yield nsname
    try:
        netns.remove(nsname)
    except OSError:
        pass


@pytest.fixture
def test_link_ifinfmsg(nsname: str) -> Generator[ifinfmsg]:
    '''Test interface ifinfmsg.

    * **Name**: test_link_ifinfmsg
    * **Scope**: function
    * **Depends**: nsname

    Create a test interface in the test netns and yield ifinfmsg. Remove
    the interface on cleanup.

    This fixture depends on **nsname**, and it means that the network
    namespace will be created automatically if you use this fixture.

    Example usage:

    .. testcode:: test_link_ifinfmsg

        def test_check_interface(nsname, test_link_ifinfmsg):
            link = test_link_ifinfmsg
            ns = ['ip', 'netns', 'exec', nsname]
            up = ['ip', 'link', 'set', 'dev', link.get('ifname'), 'up']
            subprocess.Popen(ns + up)
            # ...

    '''
    ifname = uifname()
    with IPRoute(netns=nsname) as ipr:
        ipr.link('add', ifname=ifname, kind='dummy', state='up')
        (link,) = ipr.poll(ipr.link, 'dump', ifname=ifname, timeout=5)
        yield link
        try:
            ipr.link('del', index=link.get('index'))
        except NetlinkError as e:
            if e.code != errno.ENODEV:
                raise


@pytest.fixture
def test_link(
    nsname: str, test_link_ifinfmsg: ifinfmsg
) -> Generator[TestInterface]:
    '''Test interface spec.

    * **Name**: test_link
    * **Scope**: function
    * **Depends**: nsname, test_link_ifinfmsg

    Yield `TestInterface` object for the test interface, providing
    a convenient way to access some important interface properties.
    '''
    yield TestInterface(
        index=test_link_ifinfmsg.get('index'),
        ifname=test_link_ifinfmsg.get('ifname'),
        address=test_link_ifinfmsg.get('address'),
        nsname=nsname,
    )


@pytest.fixture
def test_link_address(test_link: TestInterface) -> Generator[str]:
    '''Test interface MAC address.

    * **Name**: test_link_address
    * **Scope**: function
    * **Depends**: test_link

    Yield test interface MAC address. The network namespace and
    the test interface exist at this point.
    '''
    yield test_link.address


@pytest.fixture
def test_link_index(test_link: TestInterface) -> Generator[int]:
    '''Test interface MAC index.

    * **Name**: test_link_index
    * **Scope**: function
    * **Depends**: test_link

    Yield test interface index. The network namespace and
    the test interface exist at this point.
    '''
    yield test_link.index


@pytest.fixture
def test_link_ifname(test_link: TestInterface) -> Generator[str]:
    '''Test interface MAC ifname.

    * **Name**: test_link_ifname
    * **Scope**: function
    * **Depends**: test_link

    Yield test interface ifname. The network namespace and
    the test interface exist at this point.
    '''
    yield test_link.ifname


@pytest.fixture
def tmp_link_ifname(nsname: str) -> Generator[str]:
    '''Temporary link name.

    * **Name**: tmp_link_ifname
    * **Scope**: function
    * **Depends**: nsname

    Yield tmp link ifname, but don't create it. Try to remove
    the link on cleanup.
    '''
    ifname = uifname()
    with IPRoute(netns=nsname) as ipr:
        yield ifname
        try:
            (link,) = ipr.link('get', ifname=ifname)
            ipr.link('del', index=link.get('index'))
        except NetlinkError as e:
            if e.code != errno.ENODEV:
                raise


@pytest_asyncio.fixture
async def async_ipr(request, nsname: str) -> AsyncGenerator[AsyncIPRoute]:
    '''`AsyncIPRoute` instance.

    * **Name**: async_ipr
    * **Scope**: function
    * **Depends**: nsname

    Yield `AsyncIPRoute` instance, running within the test network namespace.
    You can provide additional keyword arguments to `AsyncIPRoute`:

    .. testcode:: async_ipr

        @pytest.mark.parametrize(
            'async_ipr',
            [
                {
                    'ext_ack': True,
                    'strict_check': True,
                },
            ],
            indirect=True
        )
        @pytest.mark.asyncio
        async def test_my_case(async_ipr):
            await async_ipr.link('set', index=1, state='up')

    '''
    kwarg = getattr(request, 'param', {})
    async with AsyncIPRoute(netns=nsname, **kwarg) as ipr:
        yield ipr


@pytest.fixture
def sync_ipr(request, nsname: str) -> Generator[IPRoute]:
    '''`IPRoute` instance.

    * **Name**: sync_ipr
    * **Scope**: function
    * **Depends**: nsname

    Yield `IPRoute` instance, running within the test network namespace.
    You can provide additional keyword arguments to `IPRoute`:

    .. testcode:: sync_ipr

        @pytest.mark.parametrize(
            'sync_ipr',
            [
                {
                    'ext_ack': True,
                    'strict_check': True,
                },
            ],
            indirect=True
        )
        def test_my_case(sync_ipr):
            sync_ipr.link('set', index=1, state='up')
    '''
    kwarg = getattr(request, 'param', {})
    with IPRoute(netns=nsname, **kwarg) as ipr:
        yield ipr


@pytest_asyncio.fixture
async def async_context(
    async_ipr: AsyncIPRoute, test_link: TestInterface
) -> AsyncGenerator[TestContext[AsyncIPRoute]]:
    '''Asynchronous TestContext.

    * **Name**: async_context
    * **Scope**: function
    * **Depends**: async_ipr, test_link

    Yield `TestContext` with `AsyncIPRoute`.
    '''
    yield TestContext[AsyncIPRoute](async_ipr, test_link)


@pytest.fixture
def sync_context(
    sync_ipr: IPRoute, test_link: TestInterface
) -> Generator[TestContext[IPRoute]]:
    '''Synchronous TestContext.

    * **Name**: sync_context
    * **Scope**: function
    * **Depends**: sync_ipr, test_link

    Yield `TestContext` with `IPRoute`.
    '''
    yield TestContext[IPRoute](sync_ipr, test_link)


@pytest.fixture
def setns_context(nsname: str) -> Generator[SetNSContext]:
    '''Set network namespace.

    * **Name**: setns_context
    * **Scope**: function
    * **Depends**: nsname

    Set a unique network namespace for the current process. Push the new
    netns onto the stack, yield the network namespace context, and pop and
    cleanup netns on exit.

    Please notice that `setns()` call affects the whole python process.
    '''
    with SetNSContext(nsname) as ctx:
        yield ctx



from collections import namedtuple

conv_map_tuple = namedtuple(
    'conv_map_tuple', 'has_attr has_netlink has_dict parser_cls'
)


class nfta_nla_parser(object):
    conv_maps = ()

    def __init__(self, **kwargs):
        for c in self.conv_maps:
            setattr(self, c.has_attr, kwargs[c.has_attr])

    def __repr__(self):
        s = ''
        for c in self.conv_maps:
            s += 'c={0}, VALUE={1}\n'.format(c, getattr(self, c.has_attr))
        return s

    @classmethod
    def from_netlink(cls, ndmsg):
        kwargs = {}
        for c in cls.conv_maps:
            if c.has_netlink is None:
                continue
            p = getattr(cls, 'cparser_' + c.parser_cls)
            nl_val = ndmsg.get_attr(c.has_netlink)
            if nl_val is None:
                kwargs[c.has_attr] = None
            else:
                kwargs[c.has_attr] = p.from_netlink(
                    ndmsg.get_attr(c.has_netlink)
                )
        return cls(**kwargs)

    def to_netlink(self):
        nla = {'attrs': []}
        for c in self.conv_maps:
            val = getattr(self, c.has_attr)
            if val is None:
                continue
            nla['attrs'].append(
                (
                    c.has_netlink,
                    getattr(self, 'cparser_' + c.parser_cls).to_netlink(val),
                )
            )
        return nla

    @classmethod
    def from_dict(cls, d):
        kwargs = {}
        for c in cls.conv_maps:
            if c.has_dict in d:
                kwargs[c.has_attr] = getattr(
                    cls, 'cparser_' + c.parser_cls
                ).from_dict(d[c.has_dict])
            else:
                kwargs[c.has_attr] = None
        return cls(**kwargs)

    def to_dict(self):
        d = {}
        for c in self.conv_maps:
            val = getattr(self, c.has_attr)
            if val is not None:
                val = getattr(self, 'cparser_' + c.parser_cls).to_dict(val)
                if val is not None:
                    d[c.has_dict] = val
        return d

    class cparser_raw(object):
        @staticmethod
        def from_netlink(val):
            return val

        @staticmethod
        def to_netlink(val):
            return val

        @staticmethod
        def from_dict(val):
            return val

        @staticmethod
        def to_dict(val):
            return val



'''

Usage::

    # Imports
    from pyroute2 import NDB, WireGuard

    IFNAME = 'wg1'

    # Create a WireGuard interface
    with NDB() as ndb:
        with ndb.interfaces.create(kind='wireguard', ifname=IFNAME) as link:
            link.add_ip('10.0.0.1/24')
            link.set(state='up')

    # Create WireGuard object
    wg = WireGuard()

    # Add a WireGuard configuration + first peer
    peer = {'public_key': 'TGFHcm9zc2VCaWNoZV9DJ2VzdExhUGx1c0JlbGxlPDM=',
            'endpoint_addr': '8.8.8.8',
            'endpoint_port': 8888,
            'persistent_keepalive': 15,
            'allowed_ips': ['10.0.0.0/24', '8.8.8.8/32']}
    wg.set(IFNAME, private_key='RCdhcHJlc0JpY2hlLEplU2VyYWlzTGFQbHVzQm9ubmU=',
           fwmark=0x1337, listen_port=2525, peer=peer)

    # Add second peer with preshared key
    peer = {'public_key': 'RCdBcHJlc0JpY2hlLFZpdmVMZXNQcm9iaW90aXF1ZXM=',
            'preshared_key': 'Pz8/V2FudFRvVHJ5TXlBZXJvR3Jvc3NlQmljaGU/Pz8=',
            'endpoint_addr': '8.8.8.8',
            'endpoint_port': 9999,
            'persistent_keepalive': 25,
            'allowed_ips': ['::/0']}
    wg.set(IFNAME, peer=peer)

    # Delete second peer
    peer = {'public_key': 'RCdBcHJlc0JpY2hlLFZpdmVMZXNQcm9iaW90aXF1ZXM=',
            'remove': True}
    wg.set(IFNAME, peer=peer)

    # Get information of the interface
    wg.info(IFNAME)

    # Get specific value from the interface
    wg.info(IFNAME)[0].get('WGDEVICE_A_PRIVATE_KEY')


NOTES:

* The `get()` method always returns iterable
* Using `set()` method only requires an interface name
* The `peer` structure is described as follow::

    struct peer_s {
        public_key:            # Base64 public key - required
        remove:                # Boolean - optional
        update_only:           # Boolean - optional
        preshared_key:         # Base64 preshared key - optional
        endpoint_addr:         # IPv4 or IPv6 endpoint - optional
        endpoint_port :        # endpoint Port - required only if endpoint_addr
        persistent_keepalive:  # time in seconds to send keep alive - optional
        replace_allowed_ips:   # Boolean - optional
        allowed_ips:           # list of CIDRs allowed - optional
    }
'''

import errno
import logging
import struct
from base64 import b64decode, b64encode
from binascii import a2b_hex
from socket import AF_INET, AF_INET6, inet_ntop, inet_pton
from time import ctime

from pyroute2.netlink import (
    NLA_F_NESTED,
    NLM_F_ACK,
    NLM_F_DUMP,
    NLM_F_REQUEST,
    genlmsg,
    nla,
)
from pyroute2.netlink.generic import (
    AsyncGenericNetlinkSocket,
    GenericNetlinkSocket,
)

# Defines from uapi/wireguard.h
WG_GENL_NAME = "wireguard"
WG_GENL_VERSION = 1
WG_KEY_LEN = 32

# WireGuard Device commands
WG_CMD_GET_DEVICE = 0
WG_CMD_SET_DEVICE = 1

# Wireguard Device attributes
WGDEVICE_A_UNSPEC = 0
WGDEVICE_A_IFINDEX = 1
WGDEVICE_A_IFNAME = 2
WGDEVICE_A_PRIVATE_KEY = 3
WGDEVICE_A_PUBLIC_KEY = 4
WGDEVICE_A_FLAGS = 5
WGDEVICE_A_LISTEN_PORT = 6
WGDEVICE_A_FWMARK = 7
WGDEVICE_A_PEERS = 8

# WireGuard Device flags
WGDEVICE_F_REPLACE_PEERS = 0x1

# WireGuard Allowed IP attributes
WGALLOWEDIP_A_UNSPEC = 0
WGALLOWEDIP_A_FAMILY = 1
WGALLOWEDIP_A_IPADDR = 2
WGALLOWEDIP_A_CIDR_MASK = 3

# WireGuard Peer flags
WGPEER_F_REMOVE_ME = 0x1
WGPEER_F_REPLACE_ALLOWEDIPS = 0x2
WGPEER_F_UPDATE_ONLY = 0x4

# Specific defines
WG_MAX_PEERS = 1000
WG_MAX_ALLOWEDIPS = 1000


class wgmsg(genlmsg):
    prefix = 'WGDEVICE_A_'

    nla_map = (
        ('WGDEVICE_A_UNSPEC', 'none'),
        ('WGDEVICE_A_IFINDEX', 'uint32'),
        ('WGDEVICE_A_IFNAME', 'asciiz'),
        ('WGDEVICE_A_PRIVATE_KEY', 'parse_wg_key'),
        ('WGDEVICE_A_PUBLIC_KEY', 'parse_wg_key'),
        ('WGDEVICE_A_FLAGS', 'uint32'),
        ('WGDEVICE_A_LISTEN_PORT', 'uint16'),
        ('WGDEVICE_A_FWMARK', 'uint32'),
        ('WGDEVICE_A_PEERS', '*wgdevice_peer'),
    )

    class wgdevice_peer(nla):
        prefix = 'WGPEER_A_'

        nla_flags = NLA_F_NESTED
        nla_map = (
            ('WGPEER_A_UNSPEC', 'none'),
            ('WGPEER_A_PUBLIC_KEY', 'parse_peer_key'),
            ('WGPEER_A_PRESHARED_KEY', 'parse_peer_key'),
            ('WGPEER_A_FLAGS', 'uint32'),
            ('WGPEER_A_ENDPOINT', 'parse_endpoint'),
            ('WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL', 'uint16'),
            ('WGPEER_A_LAST_HANDSHAKE_TIME', 'parse_handshake_time'),
            ('WGPEER_A_RX_BYTES', 'uint64'),
            ('WGPEER_A_TX_BYTES', 'uint64'),
            ('WGPEER_A_ALLOWEDIPS', '*wgpeer_allowedip'),
            ('WGPEER_A_PROTOCOL_VERSION', 'uint32'),
        )

        class parse_peer_key(nla):
            fields = (('key', '32s'),)

            def decode(self):
                nla.decode(self)
                self['value'] = b64encode(self['key'])

            def encode(self):
                self['key'] = b64decode(self['value'])
                nla.encode(self)

        @staticmethod
        def parse_endpoint(nla, *argv, **kwarg):
            family = AF_INET
            if 'data' in kwarg:
                # decoding, fetch the famliy from the NLA data
                data = kwarg['data']
                offset = kwarg['offset']
                family = struct.unpack('H', data[offset + 4 : offset + 6])[0]
            elif kwarg['value']['addr'].find(':') > -1:
                # encoding, setup family from the addr format
                family = AF_INET6

            if family == AF_INET:
                return nla.endpoint_ipv4
            else:
                return nla.endpoint_ipv6

        class endpoint_ipv4(nla):
            fields = (
                ('family', 'H'),
                ('port', '>H'),
                ('addr', '4s'),
                ('__pad', '8x'),
            )

            def decode(self):
                nla.decode(self)
                self['addr'] = inet_ntop(AF_INET, self['addr'])

            def encode(self):
                self['family'] = AF_INET
                self['addr'] = inet_pton(AF_INET, self['addr'])
                nla.encode(self)

        class endpoint_ipv6(nla):
            fields = (
                ('family', 'H'),
                ('port', '>H'),
                ('flowinfo', '>I'),
                ('addr', '16s'),
                ('scope_id', '>I'),
            )

            def decode(self):
                nla.decode(self)
                self['addr'] = inet_ntop(AF_INET6, self['addr'])

            def encode(self):
                self['family'] = AF_INET6
                self['addr'] = inet_pton(AF_INET6, self['addr'])
                nla.encode(self)

        class parse_handshake_time(nla):
            fields = (('tv_sec', 'Q'), ('tv_nsec', 'Q'))

            def decode(self):
                nla.decode(self)
                self['latest handshake'] = ctime(self['tv_sec'])

        class wgpeer_allowedip(nla):
            prefix = 'WGALLOWEDIP_A_'

            nla_flags = NLA_F_NESTED
            nla_map = (
                ('WGALLOWEDIP_A_UNSPEC', 'none'),
                ('WGALLOWEDIP_A_FAMILY', 'uint16'),
                ('WGALLOWEDIP_A_IPADDR', 'hex'),
                ('WGALLOWEDIP_A_CIDR_MASK', 'uint8'),
            )

            def decode(self):
                nla.decode(self)
                family = self.get_attr('WGALLOWEDIP_A_FAMILY')
                if family is None:
                    # Prevent when decode() is called without attrs because all
                    # datas transfered to 'value' entry.
                    #  {'attrs': [], 'value': [{'attrs' ...
                    return
                ipaddr = self.get_attr('WGALLOWEDIP_A_IPADDR')
                cidr = self.get_attr('WGALLOWEDIP_A_CIDR_MASK')
                self['addr'] = '{ipaddr}/{cidr}'.format(
                    ipaddr=inet_ntop(family, a2b_hex(ipaddr.replace(':', ''))),
                    cidr=cidr,
                )

    class parse_wg_key(nla):
        fields = (('key', '32s'),)

        def decode(self):
            nla.decode(self)
            self['value'] = b64encode(self['key'])

        def encode(self):
            self['key'] = b64decode(self['value'])
            nla.encode(self)


class AsyncWireGuard(AsyncGenericNetlinkSocket):

    async def setup_endpoint(self):
        if getattr(self.local, 'transport', None) is not None:
            return
        await super().setup_endpoint()
        await self.bind(WG_GENL_NAME, wgmsg)

    async def info(self, interface):
        msg = wgmsg()
        msg['cmd'] = WG_CMD_GET_DEVICE
        msg['attrs'].append(['WGDEVICE_A_IFNAME', interface])
        return await self.nlm_request(
            msg, msg_type=self.prid, msg_flags=NLM_F_REQUEST | NLM_F_DUMP
        )

    async def set(
        self,
        interface,
        listen_port=None,
        fwmark=None,
        private_key=None,
        peer=None,
    ):
        msg = wgmsg()
        msg['attrs'].append(['WGDEVICE_A_IFNAME', interface])

        if private_key is not None:
            self._wg_test_key(private_key)
            msg['attrs'].append(['WGDEVICE_A_PRIVATE_KEY', private_key])

        if listen_port is not None:
            msg['attrs'].append(['WGDEVICE_A_LISTEN_PORT', listen_port])

        if fwmark is not None:
            msg['attrs'].append(['WGDEVICE_A_FWMARK', fwmark])

        if peer is not None:
            self._wg_set_peer(msg, peer)

        # Message attributes
        msg['cmd'] = WG_CMD_SET_DEVICE
        msg['version'] = WG_GENL_VERSION
        msg['header']['type'] = self.prid
        msg['header']['flags'] = NLM_F_REQUEST | NLM_F_ACK
        msg['header']['pid'] = self.pid
        msg.encode()
        self.sendto(msg.data, (0, 0))
        (msg,) = [x async for x in self.get()]
        err = msg['header'].get('error', None)
        if err is not None:
            if hasattr(err, 'code') and err.code == errno.ENOENT:
                logging.error(
                    'Generic netlink protocol %s not found' % self.prid
                )
                logging.error('Please check if the protocol module is loaded')
            raise err
        return msg

    def _wg_test_key(self, key):
        try:
            if len(b64decode(key)) != WG_KEY_LEN:
                raise ValueError('Invalid WireGuard key length')
        except TypeError:
            raise ValueError('Failed to decode Base64 key')

    def _wg_set_peer(self, msg, peer):
        attrs = []
        wg_peer = [{'attrs': attrs}]
        if 'public_key' not in peer:
            raise ValueError('Peer Public key required')

        # Check public key validity
        public_key = peer['public_key']
        self._wg_test_key(public_key)
        attrs.append(['WGPEER_A_PUBLIC_KEY', public_key])

        # If peer removal is set to True
        if 'remove' in peer and peer['remove']:
            attrs.append(['WGPEER_A_FLAGS', WGPEER_F_REMOVE_ME])
            msg['attrs'].append(['WGDEVICE_A_PEERS', wg_peer])
            return

        # Set Endpoint
        if 'endpoint_addr' in peer and 'endpoint_port' in peer:
            attrs.append(
                [
                    'WGPEER_A_ENDPOINT',
                    {
                        'addr': peer['endpoint_addr'],
                        'port': peer['endpoint_port'],
                    },
                ]
            )

        # Set Preshared key
        if 'preshared_key' in peer:
            pkey = peer['preshared_key']
            self._wg_test_key(pkey)
            attrs.append(['WGPEER_A_PRESHARED_KEY', pkey])

        # Set Persistent Keepalive time
        if 'persistent_keepalive' in peer:
            keepalive = peer['persistent_keepalive']
            attrs.append(['WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL', keepalive])

        # Set Peer flags
        flags = 0
        if 'update_only' in peer and peer['update_only']:
            flags |= WGPEER_F_UPDATE_ONLY
        if 'replace_allowed_ips' in peer and peer['replace_allowed_ips']:
            flags |= WGPEER_F_REPLACE_ALLOWEDIPS
        attrs.append(['WGPEER_A_FLAGS', flags])

        # Set allowed IPs
        if 'allowed_ips' in peer:
            allowed_ips = self._wg_build_allowedips(peer['allowed_ips'])
            attrs.append(['WGPEER_A_ALLOWEDIPS', allowed_ips])

        msg['attrs'].append(['WGDEVICE_A_PEERS', wg_peer])

    def _wg_build_allowedips(self, allowed_ips):
        ret = []

        for index, ip in enumerate(allowed_ips):
            allowed_ip = []
            ret.append({'attrs': allowed_ip})

            if ip.find("/") == -1:
                raise ValueError('No CIDR set in allowed ip #{}'.format(index))

            addr, mask = ip.split('/')

            family = AF_INET if addr.find(":") == -1 else AF_INET6
            allowed_ip.append(['WGALLOWEDIP_A_FAMILY', family])
            allowed_ip.append(
                ['WGALLOWEDIP_A_IPADDR', inet_pton(family, addr)]
            )
            allowed_ip.append(['WGALLOWEDIP_A_CIDR_MASK', int(mask)])

        return ret


class WireGuard(GenericNetlinkSocket):
    async_class = AsyncWireGuard

    def info(self, interface):
        return self._run_sync_cleanup(self.asyncore.info, interface)

    def set(
        self,
        interface,
        listen_port=None,
        fwmark=None,
        private_key=None,
        peer=None,
    ):
        return self._run_with_cleanup(
            self.asyncore.set,
            interface,
            listen_port,
            fwmark,
            private_key,
            peer,
        )



'''DHCP client state machine helpers.'''

import functools
from enum import IntEnum, auto
from logging import getLogger
from typing import Any, Callable, Final, Protocol

LOG = getLogger(__name__)


class State(IntEnum):
    '''DHCP client states.

    see
    http://www.tcpipguide.com/free/t_DHCPGeneralOperationandClientFiniteStateMachine.htm
    '''

    OFF = 0
    INIT = auto()
    INIT_REBOOT = auto()
    REBOOTING = auto()
    REQUESTING = auto()
    SELECTING = auto()
    BOUND = auto()
    RENEWING = auto()
    REBINDING = auto()


# allowed transitions between states
TRANSITIONS: Final[dict[State, set[State]]] = {
    State.OFF: {State.INIT, State.INIT_REBOOT},
    State.INIT_REBOOT: {State.REBOOTING},
    State.REBOOTING: {State.INIT, State.BOUND},
    State.INIT: {State.SELECTING},
    State.SELECTING: {State.REQUESTING, State.INIT},
    State.REQUESTING: {State.BOUND, State.INIT},
    State.BOUND: {State.INIT, State.RENEWING, State.REBINDING},
    State.RENEWING: {State.BOUND, State.INIT, State.REBINDING},
    State.REBINDING: {State.BOUND, State.INIT},
}


class AsyncCallback(Protocol):
    __name__: str

    async def __call__(self, *a, **kwds: Any) -> Any:
        '''Protocol for async methods that makes mypy happy.'''


def state_guard(*states: State) -> Callable[[AsyncCallback], AsyncCallback]:
    '''Decorator that prevents a method from running

    if the associated instance is not in one of the given States.'''

    def decorator(meth: AsyncCallback) -> AsyncCallback:
        @functools.wraps(meth)
        async def wrapper(self, *args: Any, **kwargs: Any) -> None:
            if self.state not in states:
                LOG.debug(
                    'Ignoring call to %r in %s state',
                    meth.__name__,
                    self.state.name,
                )
                return
            await meth(self, *args, **kwargs)

        return wrapper

    return decorator



'''
General request and RTNL object data filters.
'''

import abc
import weakref
from collections import ChainMap


class RequestFilter(abc.ABC):
    pass


class RequestProcessor(dict):
    field_filters = tuple()
    mark = tuple()
    context = None
    combined = None
    parameters = None

    def __init__(self, context=None, prime=None, parameters=None):
        self.reset_filters()
        self.reset_mark()
        self.parameters = dict(parameters) if parameters else {}
        prime = {} if prime is None else prime
        self.context = (
            context if isinstance(context, (dict, weakref.ProxyType)) else {}
        )
        self.combined = ChainMap(self, self.context)
        if isinstance(prime, dict):
            self.update(prime)

    def __setitem__(self, key, value):
        if value is None:
            return
        new_data = self.filter(key, value)
        op = 'set'
        #
        if isinstance(new_data, tuple):
            op, new_data = new_data
        elif not isinstance(new_data, dict):
            raise ValueError('invalid new_data type')
        #
        if op == 'set':
            if key in self:
                super().__delitem__(key)
            for nkey, nvalue in new_data.items():
                if nkey in self:
                    super().__delitem__(nkey)
                super().__setitem__(nkey, nvalue)
            return
        #
        if op == 'patch':
            for nkey, nvalue in new_data.items():
                if nkey not in self:
                    self[nkey] = []
                self[nkey].extend(nvalue)
            return
        #
        raise RuntimeError('error applying new values')

    def reset_filters(self):
        self.field_filters = []

    def set_parameter(self, name, value):
        self.parameters[name] = value

    def reset_mark(self):
        self.mark = []

    def items(self):
        for key, value in super().items():
            if key not in self.mark:
                yield key, value

    def get_value(self, key, default=None, mode=None):
        for field_filter in self.field_filters:
            getter = getattr(field_filter, f'get_{key}', None)
            if getter is not None:
                return getter(self, mode)
        self.mark.append(key)
        return self.get(key, default)

    def filter(self, key, value):
        job = {key: value}
        ret = None

        for field_filter in self.field_filters:
            for k, v in tuple(job.items()):
                if hasattr(field_filter, 'key_transform'):
                    k = field_filter.key_transform(k)
                if (
                    hasattr(field_filter, 'allowed')
                    and k not in field_filter.allowed
                ):
                    return {}
                if hasattr(field_filter, 'policy') and not field_filter.policy(
                    k
                ):
                    return {}
                setter = getattr(field_filter, f'set_{k}', None)
                if setter is not None:
                    if ret is None:
                        ret = {}
                    ret.update(setter(ChainMap(self.combined, ret), v))

            if ret is not None:
                job = ret

        return ret if ret is not None else {key: value}

    def update(self, prime):
        for key, value in tuple(prime.items()):
            self[key] = value

    def add_filter(self, field_filter):
        self.field_filters.append(field_filter)
        return self

    def finalize(self):
        self.update(self)
        for field_filter in self.field_filters:
            if hasattr(field_filter, 'finalize'):
                field_filter.finalize(self.combined)
        return self



'''
Simplest example to set CAN bitrate.
'''

from pyroute2 import IPRoute

with IPRoute() as ip_route:
    # loolkup can0 interface
    idx = ip_route.link_lookup(ifname='can0')[0]
    link = ip_route.link('get', index=idx)

    # bring can0 interface down. CAN settings can be set only
    # if the interface is down
    if 'state' in link[0] and link[0]['state'] == 'up':
        ip_route.link('set', index=idx, state='down')

    # set CAN birate
    ip_route.link(
        'set', index=idx, kind='can', can_bittiming={'bitrate': 250000}
    )

    # bring can0 interface up
    ip_route.link('set', index=idx, state='up')



'''
Windows systems are not supported, but the library provides some
proof-of-concept how to build an RTNL-compatible core on top of
WinAPI calls.

Only two methods are provided so far. If you're interested in
extending the functionality, you're welcome to propose PRs.

.. warning::
    Using pyroute2 on Windows requires installing `win_inet_pton` module,
    you can use `pip install win_inet_pton`.
'''

import ctypes
import os
from socket import AF_INET

from pyroute2.common import dqn2int
from pyroute2.netlink import NLM_F_DUMP, NLM_F_MULTI, NLM_F_REQUEST, NLMSG_DONE
from pyroute2.netlink.rtnl import (
    RTM_GETADDR,
    RTM_GETLINK,
    RTM_GETNEIGH,
    RTM_GETROUTE,
    RTM_NEWADDR,
    RTM_NEWLINK,
    RTM_NEWNEIGH,
    RTM_NEWROUTE,
)
from pyroute2.netlink.rtnl.ifaddrmsg import ifaddrmsg
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
from pyroute2.netlink.rtnl.marshal import MarshalRtnl

MAX_ADAPTER_NAME_LENGTH = 256
MAX_ADAPTER_DESCRIPTION_LENGTH = 128
MAX_ADAPTER_ADDRESS_LENGTH = 8


class IP_ADDRESS_STRING(ctypes.Structure):
    pass


PIP_ADDRESS_STRING = ctypes.POINTER(IP_ADDRESS_STRING)
IP_ADDRESS_STRING._fields_ = [
    ('Next', PIP_ADDRESS_STRING),
    ('IpAddress', ctypes.c_byte * 16),
    ('IpMask', ctypes.c_byte * 16),
    ('Context', ctypes.c_ulong),
]


class IP_ADAPTER_INFO(ctypes.Structure):
    pass


PIP_ADAPTER_INFO = ctypes.POINTER(IP_ADAPTER_INFO)
IP_ADAPTER_INFO._fields_ = [
    ('Next', PIP_ADAPTER_INFO),
    ('ComboIndex', ctypes.c_ulong),
    ('AdapterName', ctypes.c_byte * (256 + 4)),
    ('Description', ctypes.c_byte * (128 + 4)),
    ('AddressLength', ctypes.c_uint),
    ('Address', ctypes.c_ubyte * 8),
    ('Index', ctypes.c_ulong),
    ('Type', ctypes.c_uint),
    ('DhcpEnabled', ctypes.c_uint),
    ('CurrentIpAddress', PIP_ADDRESS_STRING),
    ('IpAddressList', IP_ADDRESS_STRING),
    ('GatewayList', IP_ADDRESS_STRING),
    ('DhcpServer', IP_ADDRESS_STRING),
    ('HaveWins', ctypes.c_byte),
    ('PrimaryWinsServer', IP_ADDRESS_STRING),
    ('SecondaryWinsServer', IP_ADDRESS_STRING),
    ('LeaseObtained', ctypes.c_ulong),
    ('LeaseExpires', ctypes.c_ulong),
]


class IPRoute(object):
    def __init__(self, *argv, **kwarg):
        self.marshal = MarshalRtnl()
        self.target = kwarg.get('target') or 'localhost'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def clone(self):
        return self

    def close(self, code=None):
        pass

    def bind(self, *argv, **kwarg):
        pass

    def getsockopt(self, *argv, **kwarg):
        return 1024 * 1024

    def sendto_gate(self, msg, addr):
        #
        # handle incoming netlink requests
        #
        # sendto_gate() receives single RTNL messages as objects
        #
        cmd = msg['header']['type']
        flags = msg['header']['flags']
        seq = msg['header']['sequence_number']

        # work only on dump requests for now
        if flags != NLM_F_REQUEST | NLM_F_DUMP:
            return

        #
        if cmd == RTM_GETLINK:
            rtype = RTM_NEWLINK
            ret = self.get_links()
        elif cmd == RTM_GETADDR:
            rtype = RTM_NEWADDR
            ret = self.get_addr()
        elif cmd == RTM_GETROUTE:
            rtype = RTM_NEWROUTE
            ret = self.get_routes()
        elif cmd == RTM_GETNEIGH:
            rtype = RTM_NEWNEIGH
            ret = self.get_neighbours()

        #
        # set response type and finalize the message
        for r in ret:
            r['header']['type'] = rtype
            r['header']['flags'] = NLM_F_MULTI
            r['header']['sequence_number'] = seq

        #
        r = type(msg)()
        r['header']['type'] = NLMSG_DONE
        r['header']['sequence_number'] = seq
        ret.append(r)

        data = b''
        for r in ret:
            r.encode()
            data += r.data
        self._outq.put(data)
        os.write(self._pfdw, b'\0')

    def _GetAdaptersInfo(self):
        ret = {'interfaces': [], 'addresses': []}

        # prepare buffer
        buf = ctypes.create_string_buffer(15000)
        buf_len = ctypes.c_ulong(15000)
        (
            ctypes.windll.iphlpapi.GetAdaptersInfo(
                ctypes.byref(buf), ctypes.byref(buf_len)
            )
        )
        adapter = IP_ADAPTER_INFO.from_address(ctypes.addressof(buf))
        while True:
            mac = ':'.join(['%02x' % x for x in adapter.Address][:6])
            ifname = ctypes.string_at(
                ctypes.addressof(adapter.AdapterName)
            ).decode('utf-8')
            spec = {
                'index': adapter.Index,
                'attrs': (['IFLA_ADDRESS', mac], ['IFLA_IFNAME', ifname]),
            }

            msg = ifinfmsg().load(spec)
            msg['header']['target'] = self.target
            msg['header']['type'] = RTM_NEWLINK
            del msg['value']
            ret['interfaces'].append(msg)

            ipaddr = adapter.IpAddressList
            while True:
                addr = ctypes.string_at(
                    ctypes.addressof(ipaddr.IpAddress)
                ).decode('utf-8')
                mask = ctypes.string_at(
                    ctypes.addressof(ipaddr.IpMask)
                ).decode('utf-8')
                spec = {
                    'index': adapter.Index,
                    'family': AF_INET,
                    'prefixlen': dqn2int(mask),
                    'attrs': (
                        ['IFA_ADDRESS', addr],
                        ['IFA_LOCAL', addr],
                        ['IFA_LABEL', ifname],
                    ),
                }
                msg = ifaddrmsg().load(spec)
                msg['header']['target'] = self.target
                msg['header']['type'] = RTM_NEWADDR
                del msg['value']
                ret['addresses'].append(msg)
                if ipaddr.Next:
                    ipaddr = ipaddr.Next.contents
                else:
                    break

            if adapter.Next:
                adapter = adapter.Next.contents
            else:
                break
        return ret

    def dump(self, groups=None):
        for method in (
            self.get_links,
            self.get_addr,
            self.get_neighbours,
            self.get_routes,
        ):
            for msg in method():
                yield msg

    def get_links(self, *argv, **kwarg):
        '''
        Get network interfaces list::

            >>> pprint(ipr.get_links())
            [{'attrs': (['IFLA_ADDRESS', '52:54:00:7a:8a:49'],
                        ['IFLA_IFNAME',
                         '{F444467B-3549-455D-81F2-AB617C7421AB}']),
              'change': 0,
              'family': 0,
              'flags': 0,
              'header': {},
              'ifi_type': 0,
              'index': 7}]
        '''
        return self._GetAdaptersInfo()['interfaces']

    def get_addr(self, *argv, **kwarg):
        '''
        Get IP addresses::

            >>> pprint(ipr.get_addr())
            [{'attrs': (['IFA_ADDRESS', '192.168.122.81'],
                        ['IFA_LOCAL', '192.168.122.81'],
                        ['IFA_LABEL',
                         '{F444467B-3549-455D-81F2-AB617C7421AB}']),
              'family': <AddressFamily.AF_INET: 2>,
              'flags': 0,
              'header': {},
              'index': 7,
              'prefixlen': 24,
              'scope': 0}]
        '''
        return self._GetAdaptersInfo()['addresses']

    def get_neighbours(self, *argv, **kwarg):
        ret = []
        return ret

    def get_routes(self, *argv, **kwarg):
        ret = []
        return ret


class RawIPRoute(IPRoute):
    pass


class ChaoticIPRoute:
    def __init__(self, *argv, **kwarg):
        raise NotImplementedError()


class NetNS:
    def __init__(self, *argv, **kwarg):
        raise NotImplementedError()



'''
flow
++++

Flow filter supports two types of modes::
    - map
    - hash

    # Prepare a Qdisc with fq-codel
    ip.tc("add", "fq_codel", ifb0,
          parent=0x10001, handle=0x10010)

    # Create flow filter with hash mode
    # Single:
    keys = "src"
    # Multi (comma separated list of keys):
    keys = "src,nfct-src"
    ip.tc("add-filter", "flow", ifb0,
          mode="hash", keys=keys,
          divisor=1024, perturb=60,
          handle=0x10, baseclass=0x10010,
          parent=0x10001)


    # Create flow filter with map mode
    # Simple map dst with no OP:
    ip.tc("add-filter", "flow", ifb0,
    mode="map", key="dst",
    divisor=1024, handle=10
    baseclass=0x10010)

    # Same filter with xor OP:
    ops = [{"op": "xor", "num": 0xFF}]
    ip.tc("add-filter", "flow", ifb0,
    mode="map", key="dst",
    divisor=1024, handle=10
    baseclass=0x10010, ops=ops)

    # Complex one with addend OP (incl. minus support):
    ops = [{"op": "addend", "num": '-192.168.0.0'}]
    ip.tc("add-filter", "flow", ifb0,
    mode="map", key="dst",
    divisor=1024, handle=10
    baseclass=0x10010, ops=ops)

    # Example with multiple OPS:
    ops = [{"op": "and", "num": 0xFF},
           {"op": "rshift", "num": 4}]
    ip.tc("add-filter", "flow", ifb0,
    mode="map", key="dst",
    divisor=1024, handle=10
    baseclass=0x10010, ops=ops)


NOTES:
    When using `map` mode, use the keyword `key` to pass a key.
    When using `hash` mode, use the keyword `keys` to pass a key
    even if there is only one key.
    In `map` mode, the `num` parameter in `OPS` is always an
    integer unless if you use the OP `addend`, which can be
    a string IPv4 address. You can also add a minus sign at
    the beginning of the `num` value even if it is an IPv4
    address.
'''

from socket import htons

from pyroute2 import protocols
from pyroute2.netlink import nla
from pyroute2.netlink.rtnl.tcmsg.common import (
    get_tca_keys,
    get_tca_mode,
    get_tca_ops,
    tc_flow_keys,
    tc_flow_modes,
)
from pyroute2.netlink.rtnl.tcmsg.common_act import get_tca_action, tca_act_prio


def fix_request(request):
    if 'rate' in request:
        del request['rate']
    request['info'] = htons(
        request.get('protocol', protocols.ETH_P_ALL) & 0xFFFF
    ) | ((request.get('prio', 0) << 16) & 0xFFFF0000)


def get_parameters(kwarg):
    ret = {'attrs': []}
    attrs_map = (
        ('baseclass', 'TCA_FLOW_BASECLASS'),
        ('divisor', 'TCA_FLOW_DIVISOR'),
        ('perturb', 'TCA_FLOW_PERTURB'),
    )

    if kwarg.get('mode'):
        ret['attrs'].append(['TCA_FLOW_MODE', get_tca_mode(kwarg)])
        if kwarg.get('mode') == 'hash':
            ret['attrs'].append(['TCA_FLOW_KEYS', get_tca_keys(kwarg, 'keys')])

        if kwarg.get('mode') == 'map':
            ret['attrs'].append(['TCA_FLOW_KEYS', get_tca_keys(kwarg, 'key')])
            # Check for OPS presence
            if 'ops' in kwarg:
                get_tca_ops(kwarg, ret['attrs'])

    if kwarg.get('action'):
        ret['attrs'].append(['TCA_FLOW_ACT', get_tca_action(kwarg)])

    for k, v in attrs_map:
        r = kwarg.get(k, None)
        if r is not None:
            ret['attrs'].append([v, r])

    return ret


class options(nla):
    nla_map = (
        ('TCA_FLOW_UNSPEC', 'none'),
        ('TCA_FLOW_KEYS', 'tca_parse_keys'),
        ('TCA_FLOW_MODE', 'tca_parse_mode'),
        ('TCA_FLOW_BASECLASS', 'uint32'),
        ('TCA_FLOW_RSHIFT', 'uint32'),
        ('TCA_FLOW_ADDEND', 'uint32'),
        ('TCA_FLOW_MASK', 'uint32'),
        ('TCA_FLOW_XOR', 'uint32'),
        ('TCA_FLOW_DIVISOR', 'uint32'),
        ('TCA_FLOW_ACT', 'tca_act_prio'),
        ('TCA_FLOW_POLICE', 'hex'),
        ('TCA_FLOW_EMATCHES', 'hex'),
        ('TCA_FLOW_PERTURB', 'uint32'),
    )

    class tca_parse_mode(nla):
        fields = (('flow_mode', 'I'),)

        def decode(self):
            nla.decode(self)
            for key, value in tc_flow_modes.items():
                if self['flow_mode'] == value:
                    self['flow_mode'] = key
                    break

        def encode(self):
            self['flow_mode'] = self['value']
            nla.encode(self)

    class tca_parse_keys(nla):
        fields = (('flow_keys', 'I'),)

        def decode(self):
            nla.decode(self)

            keys = ''
            for key, value in tc_flow_keys.items():
                if value & self['flow_keys']:
                    keys = '{0},{1}'.format(keys, key)

            self['flow_keys'] = keys.strip(',')

        def encode(self):
            self['flow_keys'] = self['value']
            nla.encode(self)

    tca_act_prio = tca_act_prio
